"""
This module provides classes and functions for managing various assets such as buttons, selectors, and state variables in a graphical user interface.

Classes:
    Button - Add a 'name' state to a matplotlib widget button.
    StateButton - Store a number/coordinate in a button.
    ToggleButton - Add a toggle button to a matplotlib figure.
    Selector - Select points in a plot using the lasso selection widget.
    StateVariable - Manage state variables with multiple states.

    AssetContainer - Container for managing assets such as buttons, memory slots, etc.
    
    Buttons - Manager for buttons in a matplotlib figure or GUI.
    Selectors - Manager for selector objects for picking points on line2D objects.
    MemorySlots - Manager for memory slots to store and navigate positions.
    StateVariables - Manager for state variables.
"""

from __future__ import annotations

import numpy as np
from matplotlib import lines as mlines
from matplotlib import pyplot as plt
from matplotlib.path import Path as mPath
from matplotlib.widgets import Button as ButtonWidget
from matplotlib.widgets import LassoSelector as LassoSelectorWidget
from typing import Any, Callable, List, Optional, Union

from .utils import TextView


class Button(ButtonWidget):
    """Add a 'name' state to a matplotlib widget button."""

    def __init__(self, ax, name: str, **kwargs) -> None:
        super().__init__(ax, name, **kwargs)
        self.name = name


class StateButton(Button):
    """Store a number/coordinate in a button."""

    def __init__(self, ax, name: str, start_state: Any, **kwargs) -> None:
        super().__init__(ax, name, **kwargs)
        self.state = start_state  # stores something in the state


class ToggleButton(StateButton):
    """
    Add a toggle button to a matplotlib figure.

    For example usage, see PlotBrowser.
    """

    def __init__(self, ax, name: str, start_state: bool = True, **kwargs) -> None:
        super().__init__(ax, name, start_state, **kwargs)
        self.on_clicked(self.toggle)
        self.set_text()

    def set_text(self) -> None:
        """Set the text of the toggle button."""
        self.label._text = f"{self.name}={self.state}"

    def toggle(self, event=None) -> None:
        """Toggle the state of the button."""
        self.state = not self.state
        self.set_text()

    def set_state(self, state: bool) -> None:
        """Set the state of the button."""
        assert isinstance(state, bool)
        self.state = state
        self.set_text()


class Selector:
    """
    Select points in a plot using the lasso selection widget.

    Indices of selected points are stored in self.sel.

    Example:
        f, ax = plt.subplots(1, 1)
        ph, = ax.plot(np.random.rand(20))
        plt.show(block=False)
        ls = gui.Lasso(ph)
        ls.start()
        -- play around with selecting points --
        ls.stop() -> disconnects the events
    """

    def __init__(self, plot_handle: mlines.Line2D) -> None:
        """Initialize the selector with a plot handle."""
        assert isinstance(plot_handle, mlines.Line2D)
        self.plot_handle = plot_handle
        self.xdata, self.ydata = plot_handle.get_data()
        self.ax = plot_handle.axes
        (self.overlay_handle,) = self.ax.plot([], [], ".")
        self.sel = np.zeros(self.xdata.shape, dtype=bool)
        self.is_active = False

    def get_data(self) -> np.ndarray:
        """Get the data points of the plot."""
        return np.vstack((self.xdata, self.ydata)).T

    def onselect(self, verts: List[tuple]) -> None:
        """Select if not previously selected; Unselect if previously selected."""
        selected_ind = mPath(verts).contains_points(self.get_data())
        self.sel = np.logical_xor(selected_ind, self.sel)
        sel_x = list(self.xdata[self.sel])
        sel_y = list(self.ydata[self.sel])
        self.overlay_handle.set_data(sel_x, sel_y)
        plt.draw()

    def start(self, event=None) -> None:
        """Start the lasso selection."""
        self.lasso = LassoSelectorWidget(self.plot_handle.axes, self.onselect)
        self.is_active = True

    def stop(self, event=None) -> None:
        """Stop the lasso selection."""
        self.lasso.disconnect_events()
        self.is_active = False

    def toggle(self, event=None) -> None:
        """Toggle the lasso selection."""
        if self.is_active:
            self.stop(event)
        else:
            self.start(event)


class AssetContainer:
    """
    Container for assets such as a button, memoryslot, etc.

    Args:
        parent (Any): matplotlib figure, or something that has a 'figure' attribute that is a figure.
    """

    def __init__(self, parent: Any) -> None:
        self._list: List[Any] = []  # list of assets
        self.parent = parent

    def __len__(self) -> int:
        return len(self._list)

    @property
    def names(self) -> List[str]:
        return [x.name for x in self._list]

    def __getitem__(self, key: Union[int, str]) -> Any:
        """Return an asset by the name key or by position in the list."""
        if not self._has_names():
            assert isinstance(key, int)

        if isinstance(key, int) and key not in self.names:
            return self._list[key]

        return {x.name: x for x in self._list}[key]

    def _has_names(self) -> bool:
        try:
            self.names
            return True
        except AttributeError:
            return False

    def add(self, asset: Any) -> Any:
        """Add an asset to the container."""
        if hasattr(asset, "name"):
            assert asset.name not in self.names
        self._list.append(asset)
        return asset

    def __contains__(self, item: str) -> bool:
        """Check if an asset with the given name exists in the container."""
        return item in self.names


class Buttons(AssetContainer):
    """Manager for buttons in a matplotlib figure or GUI (see GenericBrowser for example)."""

    def add(
        self,
        text: str = "Button",
        action_func: Optional[Union[Callable, List[Callable]]] = None,
        pos: Optional[tuple] = None,
        w: float = 0.25,
        h: float = 0.05,
        buf: float = 0.01,
        type_: str = "Push",
        **kwargs,
    ) -> Button:
        """
        Add a button to the parent figure / object.

        If pos is provided, then w, h, and buf will be ignored.
        """
        assert type_ in ("Push", "Toggle")
        nbtn = len(self)
        if pos is None:  # start adding at the top left corner
            parent_fig = self.parent.figure
            mul_factor = 6.4 / parent_fig.get_size_inches()[0]

            btn_w = w * mul_factor
            btn_h = h * mul_factor
            btn_buf = buf
            pos = (
                btn_buf,
                (1 - btn_buf) - ((btn_buf + btn_h) * (nbtn + 1)),
                btn_w,
                btn_h,
            )

        if type_ == "Toggle":
            b = ToggleButton(plt.axes(pos), text, **kwargs)
        else:
            b = Button(plt.axes(pos), text, **kwargs)

        if action_func is not None:  # more than one can be attached
            if isinstance(action_func, (list, tuple)):
                for af in action_func:
                    b.on_clicked(af)
            else:
                b.on_clicked(action_func)

        return super().add(b)


class Selectors(AssetContainer):
    """Manager for selector objects - for picking points on line2D objects."""

    def add(self, plot_handle: mlines.Line2D) -> Selector:
        """Add a selector to the container."""
        return super().add(Selector(plot_handle))


class MemorySlots(AssetContainer):
    """Manager for memory slots to store and navigate positions."""

    def __init__(self, parent: Any) -> None:
        super().__init__(parent)
        self._list = self.initialize()
        self._memtext = None

    @staticmethod
    def initialize() -> dict:
        """Initialize memory slots."""
        return {str(k): None for k in range(1, 10)}

    def disable(self) -> None:
        """Disable memory slots."""
        self._list = {}

    def enable(self) -> None:
        """Enable memory slots."""
        self._list = self.initialize()

    def show(self, pos: str = "bottom left") -> None:
        """Show memory slot text."""
        self._memtext = TextView(self._list, fax=self.parent.figure, pos=pos)

    def update(self, key: str) -> None:
        """
        Handle memory slot updates.

        Initiate when None, go to the slot if it exists, free slot if pressed when it exists.
        key is the event.key triggered by a callback.
        """
        if self._list[key] is None:
            self._list[key] = self.parent._current_idx
            self.update_display()
        elif self._list[key] == self.parent._current_idx:
            self._list[key] = None
            self.update_display()
        else:
            self.parent._current_idx = self._list[key]
            self.parent.update()

    def update_display(self) -> None:
        """Refresh memory slot text if it is not hidden."""
        if self._memtext is not None:
            self._memtext.update(self._list)

    def hide(self) -> None:
        """Hide the memory slot text."""
        if self._memtext is not None:
            self._memtext._text.remove()
        self._memtext = None

    def is_enabled(self) -> bool:
        """Check if memory slots are enabled."""
        return bool(self._list)


class StateVariable:
    """Manage state variables with multiple states."""

    def __init__(self, name: str, states: list) -> None:
        self.name = name
        self.states = list(states)
        self._current_state_idx = 0

    @property
    def current_state(self) -> Any:
        """Get the current state."""
        return self.states[self._current_state_idx]

    def n_states(self) -> int:
        """Get the number of states."""
        return len(self.states)

    def cycle(self) -> None:
        """Cycle to the next state."""
        self._current_state_idx = (self._current_state_idx + 1) % self.n_states()

    def cycle_back(self) -> None:
        """Cycle to the previous state."""
        self._current_state_idx = (self._current_state_idx - 1) % self.n_states()

    def set_state(self, state: Union[int, str]) -> None:
        """Set the state."""
        if isinstance(state, int):
            assert 0 <= state < self.n_states()
            self._current_state_idx = state
        if isinstance(state, str):
            assert state in self.states
            self._current_state_idx = self.states.index(state)


class StateVariables(AssetContainer):
    """Manager for state variables."""

    def __init__(self, parent: Any) -> None:
        super().__init__(parent)
        self._text = None

    def asdict(self) -> dict:
        """Return state variables as a dictionary."""
        return {x.name: x.states for x in self._list}

    def add(self, name: str, states: list) -> StateVariable:
        """Add a state variable to the container."""
        assert name not in self.names
        return super().add(StateVariable(name, states))

    def _get_display_text(self) -> List[str]:
        """Get the display text for state variables."""
        return ["State variables:"] + [
            f"{x.name} - {x.current_state}" for x in self._list
        ]

    def show(self, pos: str = "bottom right", fax=None) -> None:
        """Show state variables text."""
        if fax is None:
            fax = self.parent.figure
        self._text = TextView(self._get_display_text(), fax=fax, pos=pos)

    def update_display(self, draw: bool = True) -> None:
        """Update the display of state variables."""
        self._text.update(self._get_display_text())
        if draw:
            plt.draw()



import numpy as np
from datanavigator import lucas_kanade, lucas_kanade_rstc
from decord import VideoReader
from matplotlib import pyplot as plt
from datanavigator import get_example_video


def test_lucas_kanade_rstc():
    vname = get_example_video()
    video = VideoReader(vname)
    start_frame = 35
    end_frame = 50

    start_points = [[153.81, 195.34], [231.90, 209.27]]
    end_points = [[166.24, 166.74], [246.63, 181.54]]

    forward_path = lucas_kanade(
        video, start_frame, end_frame, start_points, mode="full"
    )
    reverse_path = lucas_kanade(video, end_frame, start_frame, end_points, mode="full")

    rstc_path = lucas_kanade_rstc(
        video, start_frame, end_frame, start_points, end_points
    )

    plt.figure()
    plt.plot(
        np.array(
            [
                forward_path[:, 0, 0],
                np.flip(reverse_path, 0)[:, 0, 0],
                rstc_path[:, 0, 0],
            ]
        ).T
    )
    plt.legend(["Forward", "Reverse", "RSTC"])
    plt.show(block=False)

    n_frames = end_frame - start_frame + 1
    n_points = len(start_points)
    assert forward_path.shape == (n_frames, n_points, 2)

    direct_prediction = lucas_kanade(
        video, end_frame, start_frame, end_points, mode="direct"
    )
    assert direct_prediction.shape == (1, n_points, 2)

    return forward_path, reverse_path, rstc_path



"""
Module for browsing and visualizing time series data.

Classes:
    SignalBrowser: A browser for navigating through an array of `pysampled.Data` elements or 2D arrays.
"""

from __future__ import annotations

import pysampled
from matplotlib import pyplot as plt
from typing import Callable, Optional

from .core import GenericBrowser


class SignalBrowser(GenericBrowser):
    """
    Browse an array of pysampled.Data elements, or 2D arrays.
    """

    def __init__(
        self,
        plot_data: list[pysampled.Data],
        titlefunc: Optional[Callable] = None,
        figure_handle: Optional[plt.Figure] = None,
        reset_on_change: bool = False,
    ) -> None:
        """
        Initialize the SignalBrowser.

        Args:
            plot_data (list): List of data objects to browse.
            titlefunc (callable, optional): Function to generate titles. Defaults to None.
            figure_handle (plt.Figure, optional): Matplotlib figure handle. Defaults to None.
            reset_on_change (bool, optional): Whether to reset the view on change. Defaults to False.
        """
        super().__init__(figure_handle)

        self._ax = self.figure.subplots(1, 1)
        this_data = plot_data[0]
        if isinstance(this_data, pysampled.Data):
            self._plot = self._ax.plot(this_data.t, this_data())
        else:
            (self._plot,) = self._ax.plot(this_data)

        self.data = plot_data
        if titlefunc is None:
            self.titlefunc = lambda s: getattr(
                s.data[s._current_idx], "name", f"Plot number {s._current_idx}"
            )
        else:
            self.titlefunc = titlefunc

        self.reset_on_change = reset_on_change
        # initialize
        self.set_default_keybindings()
        self.buttons.add(
            text="Auto limits",
            type_="Toggle",
            action_func=self.update,
            start_state=False,
        )
        plt.show(block=False)
        self.update()

    def update(self, event=None):
        """
        Update the browser.

        Args:
            event (optional): Event that triggered the update. Defaults to None.
        """
        this_data = self.data[self._current_idx]
        if isinstance(this_data, pysampled.Data):
            data_to_plot = this_data.split_to_1d()
            for plot_handle, this_data_to_plot in zip(self._plot, data_to_plot):
                plot_handle.set_data(this_data_to_plot.t, this_data_to_plot())
        else:
            self._plot.set_ydata(this_data)
        self._ax.set_title(self.titlefunc(self))
        if (
            "Auto limits" in self.buttons and self.buttons["Auto limits"].state
        ):  # is True
            self.reset_axes()
        plt.draw()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'datanavigator'
copyright = '2024-present, Praneeth Namburi'
author = 'Praneeth Namburi'
release = '1.0.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.duration",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",
    "sphinxcontrib.youtube",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "collapse_navigation": False,
}
html_static_path = []

autodoc_member_order = "bysource"
napoleon_use_ivar = True

from sphinx.ext.autodoc import between


def setup(app):
    # Register a sphinx.ext.autodoc.between listener to ignore everything
    # between lines that contain the word IGNORE
    app.connect("autodoc-process-docstring", between("^.*IGNORE.*$", exclude=True))
    return app



"""
This module provides functions to set and get the paths for the clip and cache folders.

The clip folder is used to store video clips, for example, when using VideoBrowser. 
The cache folder is used by the :py:mod:`datanavigator.examples` to write json file containing marked events.
"""

import os
import logging

# Configure logging
logging.basicConfig(level=logging.INFO)

# Use environment variables for default paths
CLIP_FOLDER: str = os.getenv("CLIP_FOLDER", "C:\\data\\_clipcollection")
CACHE_FOLDER: str = os.getenv("CACHE_FOLDER", "C:\\data\\_cache")


def set_clip_folder(folder: str) -> None:
    """Set the path for storing video clips."""
    if not os.path.exists(folder):
        raise ValueError(f"The provided folder path does not exist: {folder}")

    global CLIP_FOLDER
    CLIP_FOLDER = folder
    logging.info(f"Clip folder set to: {CLIP_FOLDER}")

    global CACHE_FOLDER
    if not os.path.exists(CACHE_FOLDER):
        logging.info(
            "Setting the cache folder to be the same as the clip folder. To change, use set_cache_folder(<folder_name>)."
        )
        CACHE_FOLDER = folder


def get_clip_folder() -> str:
    """Get the current path of the clip folder."""
    return CLIP_FOLDER


def set_cache_folder(folder: str) -> None:
    """Set the path for the cache folder."""
    if not os.path.exists(folder):
        raise ValueError(f"The provided folder path does not exist: {folder}")

    global CACHE_FOLDER
    CACHE_FOLDER = folder
    logging.info(f"Cache folder set to: {CACHE_FOLDER}")


def get_cache_folder() -> str:
    """Get the current path of the cache folder."""
    return CACHE_FOLDER



from datanavigator.signals import SignalBrowser
from tests.conftest import simulate_key_press, press_browser_button


def test_signal_browser_init_with_pysampled_data(signal_list, mock_figure):
    """Test initialization with pysampled.Data objects."""
    browser = SignalBrowser(plot_data=signal_list, figure_handle=mock_figure)

    assert browser.data == signal_list
    assert callable(browser.titlefunc)
    assert browser.reset_on_change is False
    assert browser.titlefunc(browser) == "Plot number 0"
    event = simulate_key_press(browser.figure, key="right")
    browser(event)
    assert browser._current_idx == 1
    assert browser.titlefunc(browser) == "Plot number 1"


def test_signal_browser_init_with_2d_array(mock_figure):
    """Test initialization with 2D array."""
    plot_data = [[10, 20, 30, 40], [1, 5, 4, 3]]
    browser = SignalBrowser(plot_data=plot_data, figure_handle=mock_figure)

    assert browser.titlefunc(browser) == "Plot number 0"
    event = simulate_key_press(browser.figure, key="right")
    browser(event)
    assert browser._current_idx == 1
    assert browser.titlefunc(browser) == "Plot number 1"


def test_signal_browser_init_with_titlefunc(signal_list, mock_figure):
    """Test initialization with a custom title function."""
    custom_titlefunc = lambda s: "Custom Title"
    browser = SignalBrowser(
        plot_data=signal_list, titlefunc=custom_titlefunc, figure_handle=mock_figure
    )

    assert browser.titlefunc == custom_titlefunc
    assert browser.titlefunc(browser) == "Custom Title"


def test_signal_browser_init_with_reset_on_change(signal_list, mock_figure):
    """Test initialization with reset_on_change set to True."""
    browser = SignalBrowser(
        plot_data=signal_list, reset_on_change=True, figure_handle=mock_figure
    )

    assert browser.reset_on_change is True


def test_signal_browser_buttons_added(signal_list, mock_figure):
    """Test that buttons are added during initialization and simulate a mouse click on the 'Right' button."""
    browser = SignalBrowser(plot_data=signal_list, figure_handle=mock_figure)

    # Add the "Right" button
    browser.buttons.add(
        text="Right",
        type_="Push",
        action_func=(lambda s, event: s.increment(step=1)).__get__(browser),
    )
    browser.update()

    # Add the "Right" button
    browser.buttons.add(
        text="Left",
        type_="Push",
        action_func=(lambda s, event: s.decrement(step=1)).__get__(browser),
    )
    browser.update()

    # Assert that the increment method was called
    press_browser_button(browser.buttons["Right"])
    assert browser._current_idx == 1
    assert browser.titlefunc(browser) == "Plot number 1"

    assert browser.buttons["Auto limits"].state is False
    press_browser_button(browser.buttons["Auto limits"])
    assert browser.buttons["Auto limits"].state is True

    press_browser_button(browser.buttons["Left"])
    assert browser._current_idx == 0
    assert browser.titlefunc(browser) == "Plot number 0"



import pytest
import numpy as np
import matplotlib.pyplot as plt

from datanavigator.plots import PlotBrowser
from tests.conftest import simulate_key_press, press_browser_button


@pytest.fixture(scope="module")
def figure_with_two_subplots():
    # Set up: create the figure and axis
    fig, ax_all = plt.subplots(2, 1, figsize=(5, 10))
    yield fig, ax_all
    # Teardown: close the figure
    plt.close(fig)


def plotter(figure_with_two_subplots):
    fig, ax_all = figure_with_two_subplots

    def setup(data):
        plot_handles = {}
        plot_handles["ax1"] = ax_all[0]
        plot_handles["ax2"] = ax_all[1]
        plot_handles["ax"] = ax_all
        plot_handles["figure"] = fig
        (plot_handles["p1"],) = ax_all[0].plot([], [], label="p1")
        (plot_handles["p2"],) = ax_all[1].plot([], [], label="p2")
        return plot_handles

    def update(data, plot_handles):
        plot_handles["p1"].set_data(data.t, data())
        plot_handles["p2"].set_data(data.t, data.lowpass(2)())

    return setup, update


def plotter_no_setup(figure_with_two_subplots):
    fig, ax_all = figure_with_two_subplots

    def update(data, figure_handle):
        ax_all[0].plot(data.t, data())
        ax_all[1].plot(data.t, data.lowpass(2)())

    return update


def test_plot_browser(signal_list, figure_with_two_subplots):
    pb = PlotBrowser(signal_list, plotter(figure_with_two_subplots))
    assert "p1" in pb.plot_handles
    event = simulate_key_press(pb.figure, "right")
    assert pb._current_idx == 0
    pb(event)
    assert pb._current_idx == 1
    assert np.allclose(pb.get_current_data()(), signal_list[1]())

    assert pb.buttons["Auto limits"].state is False
    press_browser_button(pb.buttons["Auto limits"])
    assert pb.buttons["Auto limits"].state is True
    pb.update()


def test_plot_browser_no_setup(signal_list, figure_with_two_subplots):
    pb = PlotBrowser(signal_list, plotter_no_setup(figure_with_two_subplots))
    assert pb.plot_handles is None
    assert pb.setup_func is None
    event = simulate_key_press(pb.figure, "right")
    assert pb._current_idx == 0
    pb(event)
    assert pb._current_idx == 1
    assert np.allclose(pb.get_current_data()(), signal_list[1]())



import pytest
import pysampled

import matplotlib.pyplot as plt
from matplotlib.backend_bases import KeyEvent
from matplotlib.backend_bases import MouseEvent

import datanavigator


@pytest.fixture
def matplotlib_figure():
    # used by test_events and test_core
    # Set up: create the figure and axis
    fig, ax = plt.subplots()
    ax.plot([0, 1], [0, 1])
    yield fig, ax
    # Tear down: close the figure
    plt.close(fig)


@pytest.fixture
def mock_figure():
    """Fixture to mock a matplotlib figure."""
    fig = plt.figure()
    yield fig
    plt.close(fig)


@pytest.fixture(autouse=True)
def close_figures():
    yield
    plt.close("all")


@pytest.fixture
def signal_list():
    """Fixture to mock pysampled.Data objects."""
    return [
        pysampled.generate_signal("white_noise"),
        pysampled.generate_signal("sine_wave"),
        pysampled.generate_signal("three_sine_waves"),
    ]


@pytest.fixture(scope="function", autouse=True)
def setup_folders(tmp_path):
    """Ensure clean cache and clip folders for each test."""
    curr_clip_dir = datanavigator.get_clip_folder()
    curr_cache_dir = datanavigator.get_cache_folder()
    clip_dir = tmp_path / "clips"
    cache_dir = tmp_path / "cache"
    clip_dir.mkdir()
    cache_dir.mkdir()
    datanavigator.set_clip_folder(str(clip_dir))
    datanavigator.set_cache_folder(str(cache_dir))
    yield str(clip_dir), str(cache_dir)
    # Teardown
    datanavigator.set_clip_folder(curr_clip_dir)
    datanavigator.set_cache_folder(curr_cache_dir)
    # Rest of the teardown is handled by tmp_path fixture


def simulate_key_press(figure, key="a", **kwargs):
    """
    Simulate a key press event on the given axis.

    Args:
        fax (tuple): A tuple containing the figure and axis (fig, ax).
        key (str, optional): The key to press. Defaults to 'a'.
    """
    # Create a KeyEvent
    event = KeyEvent(
        name="key_press_event",
        canvas=figure.canvas,
        key=key,
        guiEvent=None,
    )
    for k, v in kwargs.items():
        setattr(event, k, v)
    return event


def simulate_key_press_at_xy(fax, key="1", xdata=0.5, ydata=0.5):
    """
    Simulate a key press event on the given axis with the mouse positioned
    at a specific data coordinate. Adds 'xdata' and 'ydata' attributes
    to the event object.

    Args:
        fax (tuple): A tuple containing the figure and axis (fig, ax).
        key (str, optional): The key to press. Defaults to '1'.
        xdata (float, optional): The x-coordinate in data space. Defaults to 0.5.
        ydata (float, optional): The y-coordinate in data space. Defaults to 0.5.
    """
    fig, ax = fax
    # Convert the data coordinates to canvas coordinates
    x_pixel, y_pixel = ax.transData.transform((xdata, ydata))
    # Create a KeyEvent with pixel position information
    event = KeyEvent(
        name="key_press_event",
        canvas=fig.canvas,
        key=key,
        x=x_pixel,
        y=y_pixel,
        guiEvent=None,
    )
    # Manually add the data coordinates to the event object
    event.xdata = xdata
    event.ydata = ydata
    # Add the axes attribute, similar to MouseEvent
    event.inaxes = ax
    return event


def simulate_mouse_click(fax, xdata=0.5, ydata=0.5, button=1):
    """
    Simulate a mouse click event on the given axis.

    Args:
        ax (matplotlib.axes.Axes): The axis to click on.
        xdata (float): The x-coordinate of the click.
        ydata (float): The y-coordinate of the click.
        button (int, optional): The mouse button to use (1 for left, 2 for middle, 3 for right). Defaults to 1.
    """
    fig, ax = fax
    # Create a MouseEvent
    event = MouseEvent(
        name="button_press_event",
        canvas=ax.figure.canvas,
        x=ax.transData.transform((xdata, ydata))[0],
        y=ax.transData.transform((xdata, ydata))[1],
        button=button,
        key=None,
        step=0,
        dblclick=False,
        guiEvent=None,
    )
    return event


def press_browser_button(button: datanavigator.Button):
    """
    Simulates a mouse click on a given button in a Matplotlib-based datanavigator GUI.
    This function triggers both a button press and a button release event
    on the specified button, effectively simulating a full mouse click.

    Args:
        button (datanavigator.Button): The button object to be clicked.
            It must have an `ax` attribute representing the Matplotlib Axes
            associated with the button.

    Raises:
        AttributeError: If the `button` object does not have the required `ax` attribute.
        ValueError: If the button's Axes or its associated canvas is not properly configured.
    """
    if not hasattr(button, "ax"):
        raise AttributeError(
            "The button object must have an 'ax' attribute representing its Matplotlib Axes."
        )

    button_ax = button.ax
    if button_ax is None or button_ax.figure is None or button_ax.figure.canvas is None:
        raise ValueError(
            "The button's Axes or its associated canvas is not properly configured."
        )

    # Calculate the center position of the button in canvas coordinates
    bbox = button_ax.get_position()
    x = bbox.x0 + 0.5 * bbox.width
    y = bbox.y0 + 0.5 * bbox.height
    canvas = button_ax.figure.canvas
    canvas_width, canvas_height = canvas.get_width_height()

    # Convert normalized coordinates to canvas coordinates
    canvas_x = canvas_width * x
    canvas_y = canvas_height * y

    # Create and process a button press event
    press_event = MouseEvent(
        name="button_press_event",
        canvas=canvas,
        x=canvas_x,
        y=canvas_y,
        button=1,
        key=None,
    )
    canvas.callbacks.process("button_press_event", press_event)

    # Create and process a button release event
    release_event = MouseEvent(
        name="button_release_event",
        canvas=canvas,
        x=canvas_x,
        y=canvas_y,
        button=1,
        key=None,
    )
    canvas.callbacks.process("button_release_event", release_event)



"""
Module for browsing and visualizing data using customizable plotting functions.

Classes:
    PlotBrowser: A browser for navigating through a list of data objects and visualizing them using custom plotting functions.
"""

from __future__ import annotations
from typing import Union, Tuple, Any

from matplotlib import pyplot as plt

from .core import GenericBrowser


class PlotBrowser(GenericBrowser):
    """
    Takes a list of data, and a plotting function (or a pair of setup
    and update functions) that parses each of the elements in the array.
    Assumes that the plotting function is going to make one figure.
    """

    def __init__(
        self,
        plot_data: list,
        plot_func: Union[Tuple[callable, callable], callable],
        figure_handle: plt.Figure = None,
        **plot_kwargs,
    ):
        """
        Initialize the PlotBrowser.

        Args:
            plot_data (list): List of data objects to browse.
            plot_func (callable): Plotting function or a tuple of (setup_func, update_func).

                plot_func can be a tuple (setup_func, update_func), or just one plotting function - update_func
                If only one function is supplied, figure axes will be cleared on each update.
                setup_func takes:
                    the first element in plot_data list as its first input
                    keyword arguments (same as plot_func)
                setup_func outputs:
                    **dictionary** of plot handles that goes as the second input to update_func
                    plot_handles["figure"] is the figure handle to be used for plotting, and is required

                update_func is a plot-refreshing function that accepts 3 inputs:
                    an element in the plot_data list as its first input
                    output of the setup_func if it exists, or a figure handle on which to plot
                    keyword arguments
            figure_handle (plt.Figure, optional): Matplotlib figure handle. Defaults to None. Ideally, this is handled by the setup function.
            **plot_kwargs: Additional keyword arguments to pass to the plotting function.
        """
        self.data = plot_data  # list where each element serves as input to plot_func
        self.plot_kwargs = plot_kwargs

        if isinstance(plot_func, tuple):
            assert len(plot_func) == 2
            self.setup_func, self.plot_func = plot_func
            self.plot_handles = self.setup_func(self.data[0], **self.plot_kwargs)
            assert "figure" in self.plot_handles
            figure_handle = self.plot_handles["figure"]
        else:
            self.setup_func, self.plot_func = None, plot_func
            self.plot_handles = None
            figure_handle = figure_handle

        # setup
        super().__init__(figure_handle)

        # initialize
        self.set_default_keybindings()
        self.buttons.add(
            text="Auto limits",
            type_="Toggle",
            action_func=self.update,
            start_state=False,
        )
        self.memoryslots.show()

        # if an inherited class is accessing this, then don't run the update function here
        if self.__class__.__name__ == "PlotBrowser":
            self.update()
            self.reset_axes()
            plt.show(block=False)

        # add selectors after drawing!
        if self.plot_handles is not None:
            self.add_selectors()

    def add_selectors(self):
        try:
            s0 = self.selectors.add(list(self.plot_handles.values())[0])
            self.buttons.add(
                text="Selector 0",
                type_="Toggle",
                action_func=s0.toggle,
                start_state=s0.is_active,
            )
        except AssertionError:
            print("Unable to add selectors")

    def get_current_data(self):
        """
        Data getter. Plotting data is a list of objects that are being
        plotted one at a time. This function returns the current object.
        """
        return self.data[self._current_idx]

    def update(self, event: Any = None):
        """Update the browser."""
        if self.setup_func is None:
            self.figure.clear()  # redraw the entire figure contents each time, NOT recommended
            self.memoryslots.show()
            self.plot_func(self.get_current_data(), self.figure, **self.plot_kwargs)
        else:
            self.memoryslots.update_display()
            self.plot_func(
                self.get_current_data(), self.plot_handles, **self.plot_kwargs
            )
        if self.buttons["Auto limits"].state:  # is True
            self.reset_axes()
        super().update(event)
        plt.draw()



"""
This module provides classes and functions for managing events and event data.

Classes:
    :py:class:`EventData` - Manage the data from one event type in one trial.
    :py:class:`Event` - Manage selection of a sequence of events.
    :py:class:`Events` - Manager for event objects.
"""

from __future__ import annotations

import functools
import json
import os
from copy import deepcopy
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import portion as P
from matplotlib import pyplot as plt

from . import utils
from .assets import AssetContainer

PLOT_COLORS = plt.rcParams["axes.prop_cycle"].by_key()["color"]


class _PNInterval(P.Interval):
    """Extended portion Interval class with additional properties."""

    @property
    def atomic_durations(self) -> List[float]:
        """List of atomic durations in the event. In this case, an atom is one interval with start and end times.

        Returns:
            List[float]: Atomic durations.
        """
        return [xi.upper - xi.lower for xi in self]

    @property
    def duration(self) -> float:
        """Total duration (sum of atomic durations) of the interval.

        Returns:
            float: Total duration.
        """
        return sum(self.atomic_durations)

    @property
    def fraction(self) -> float:
        """Fractional duration relative to the enclosure.

        Returns:
            float: Fractional duration.
        """
        return self.duration / self.enclosure.duration


portion = P.create_api(_PNInterval)


class EventData:
    """
    Manage the data from one event type in one trial.

    Attributes:
        default (list): Default events created by an algorithm.
        added (list): Manually added events. Note that if an 'added' point is removed, then it will simply be deleted. There will be no record of it.
        removed (list): Events removed from the default list.
        tags (list): Tags associated with the events.
        algorithm_name (str): Name of the algorithm used to generate the default list.
        params (dict): Parameters used to generate the default list.
    """

    def __init__(
        self,
        default: Optional[List] = None,
        added: Optional[List] = None,
        removed: Optional[List] = None,
        tags: Optional[List] = None,
        algorithm_name: str = "",
        params: Optional[Dict] = None,
    ) -> None:
        def _to_list(x: Any) -> List:
            return [] if x is None else x

        def _1d_to_2d(x: List) -> List:
            v = np.asarray(x)
            if (
                v.ndim == 1
            ):  # passing in a list events of size 1, e.g., [1, 2, 3, 4]. Turn it into [[1], [2], [3], [4]]
                v = v[:, np.newaxis]
                return [list(vi) for vi in v]
            return x

        self.default = _1d_to_2d(_to_list(default))
        self.added = _1d_to_2d(_to_list(added))
        self.removed = _1d_to_2d(_to_list(removed))
        self.tags = _to_list(tags)
        self.algorithm_name = algorithm_name
        self.params = params if params is not None else {}

        # assert uniform size of events, 0 when there are no events
        assert len(set([len(x) for x in self.get_times()])) in (
            0,
            1,
        ), "All events must have the same size."

    def asdict(self) -> Dict:
        """Convert the event data to a dictionary."""
        return dict(
            default=self.default,
            added=self.added,
            removed=self.removed,
            tags=self.tags,
            algorithm_name=self.algorithm_name,
            params=self.params,
        )

    def __len__(self) -> int:
        """Return the number of events."""
        return len(self.get_times())

    def get_size(self) -> int:
        """Return the size of the envent, for example, 2 for start and stop events."""
        size_set = set([len(x) for x in self.get_times()])
        if len(size_set) == 1:
            return size_set.pop()
        raise ValueError("All events must have the same size.")

    def get_times(self) -> List:
        """Get the times of all events."""
        x = self.default + self.added
        x.sort()
        return x

    def to_portions(self) -> "portion.Interval":
        """Convert the event times to portions."""
        return functools.reduce(
            lambda a, b: a | b,
            [portion.closed(*interval_limits) for interval_limits in self.get_times()],
            portion.empty(),
        )

    def __and__(self, other: EventData) -> EventData:
        """Return the intersection of two EventData objects."""
        try:
            x = self.to_portions() & other.to_portions()
        except TypeError:
            raise TypeError(
                "Intersection is only implemented for size-2 events (start and stop times)."
            )
        return EventData(
            default=[[xi.lower, xi.upper] for xi in x if xi.upper != xi.lower]
        )  # only return non-zero duration events

    def __contains__(self, item: Any) -> bool:
        """Check if an item is in the event data."""
        return item in self.to_portions()

    @staticmethod
    def _process_inp(other: Union["portion.Interval", Tuple]) -> portion.Interval:
        """Process the input to ensure it is an interval."""
        if other.__class__.__name__ == EventData.__name__:
            return other.to_portions()
        if not isinstance(other, portion.Interval):
            assert len(other) == 2
            other = portion.closed(*other)
        return other

    def overlap_duration(self, other: Union["portion.Interval", Tuple]) -> float:
        """Calculate the duration of overlap with another interval."""
        other = self._process_inp(other)
        return (self.to_portions() & other).duration


class Event:
    """
    Manage selection of a sequence of events (of length >= 1).

    Attributes:
        name (str): Name of the event.
        size (int): Length of the sequence.
        fname (str): File name to load and save events.
        data_id_func (Callable): Function to get the current data ID from the parent UI.
        color (str): Color of the event. Defaults to a random color.
        pick_action (str): Action to take when picking an event ('overwrite' or 'append'). Use overwrite if there can only be one sequence per 'signal'. If there can be multiple, use 'append'. Defaults to overwrite
        ax_list (List): List of axes on which to show the event.
        win_remove (Tuple): Window relative to the mouse position to search for removing an event. Defaults to (-0.1, 0.1).
        win_add (Tuple): Window relative to the mouse position to search for adding an event. Defaults to (-0.25, 0.25).
        data_func (Callable): Function to process the data.
        plot_kwargs (Dict): Keyword arguments for plotting. Commonly used keys are 'display_type' (line or fill) and 'alpha' to set the transparency level.
    """

    def __init__(
        self,
        name: str,
        size: int,
        fname: str,
        data_id_func: Optional[Callable] = lambda: None,
        color: Union[str, int] = "random",
        pick_action: str = "overwrite",
        ax_list: Optional[List] = None,
        win_remove: Tuple[float, float] = (-0.1, 0.1),
        win_add: Tuple[float, float] = (-0.25, 0.25),
        data_func: Callable = float,
        **plot_kwargs,
    ) -> None:
        self.name = name
        assert isinstance(size, int) and size > 0
        self.size = size
        self.fname = str(fname)
        self.data_id_func = data_id_func
        if isinstance(color, int):
            color = PLOT_COLORS[color]
        elif color == "random":
            color = np.random.choice(PLOT_COLORS)
        self.color = color
        assert pick_action in ("overwrite", "append")
        self.pick_action = pick_action

        self._buffer = []
        _, self._data = self.load()

        if ax_list is None:
            ax_list = []
        self.ax_list = ax_list
        self.plot_handles = []

        self.win_remove = win_remove
        self.win_add = win_add
        self.plot_kwargs = plot_kwargs
        self._hide = False
        self.data_func = data_func

    @classmethod
    def _from_existing_file(
        cls, fname: str, data_id_func: Optional[Callable] = lambda: None
    ) -> Event:
        """Create an Event object by reading an existing JSON file."""
        h, _ = cls._read_json_file(fname)
        return cls(
            h["name"],
            h["size"],
            fname,
            data_id_func,
            h["color"],
            h["pick_action"],
            None,
            h["win_remove"],
            h["win_add"],
            **h["plot_kwargs"],
        )

    @classmethod
    def from_file(cls, fname: str, **kwargs) -> Event:
        """
        Create an empty events file with the given file name (fname) and any parameters.
        Assigns best-guess defaults.
        """
        if not os.path.exists(fname):
            kwargs["name"] = kwargs.get("name", Path(fname).stem)
            kwargs["size"] = kwargs.get("size", 1)
            kwargs["data_id_func"] = kwargs.get("data_id_func", lambda: None)

            ret = cls(fname=fname, **kwargs)
            ret.save()
            return ret
        return cls._from_existing_file(fname, kwargs.get("data_id_func", lambda: None))

    @classmethod
    def from_data(
        cls,
        data: Dict,
        name: str = "Event",
        fname: str = "",
        overwrite: bool = False,
        **kwargs,
    ) -> Event:
        """
        Create an event file by filling in the 'default' events extracted by an algorithm.

        Args:
            data (dict): Data to create the event file.
            name (str): Name of the event.
            fname (str): File name to save the event.
            overwrite (bool): Whether to overwrite the existing file.
            **kwargs: Additional keyword arguments.
                tags, algorithm_name, and params will be passed to :py:class:`EventData`.
                All other keyword arguments will be passed to :py:class:`Event`.

        Returns:
            Event: The created Event object.
        """
        data = deepcopy(data)

        algorithm_info = dict(
            tags=kwargs.pop("tags", []),
            algorithm_name=kwargs.pop("algorithm_name", ""),
            params=kwargs.pop("params", {}),
        )

        for key, val in data.items():
            if not isinstance(val, EventData):
                data[key] = EventData(default=val, **algorithm_info)

        size = []
        for key, val in data.items():
            # when there are no events, size cannot be inferred for that trial. Note that this process will fail if there are no events in ANY of the trials. size has to be passed in with kwargs.
            if len(val) > 0:
                size.append(data[key].get_size())

        if not size:
            # if there were no events in the data that was passed!
            assert "size" in kwargs
            size = kwargs["size"]
        else:
            size = list(set(size))
            assert len(size) == 1  # make sure all the events are of the same type
            size = size[0]

        if "size" in kwargs:
            assert kwargs["size"] == size
            del kwargs["size"]

        ret = cls(name, size, fname, **kwargs)
        if ret.pick_action == "overwrite":
            for key in data:
                data[key].default = data[key].default[-1:]  # keep only the last one
        ret._data = data

        # If the file path does not exists or is not creatable, then return the data that was passed in without saving it
        if not utils.is_path_exists_or_creatable(str(fname)):
            return ret

        # Save the file if it does not exist or if the mandate is to overwrite the file
        if (not os.path.exists(fname)) or overwrite:
            ret.save()
            return ret

        # If the file exists and we decided not to overwrite, then append new data to the file
        assert os.path.exists(fname) and (not overwrite)
        ret_existing = cls.from_file(fname, **kwargs)
        new_keys = set(ret._data.keys()) - set(ret_existing._data.keys())

        # If there is new data, then add it to the event file
        # Whether there are new keys or not, the data in the existing saved file supercedes the data that was passed in
        print(f"Appending new data to the event file {fname}:")
        print(new_keys)
        ret_existing._data = {**ret._data, **ret_existing._data}
        ret_existing.save()

        return ret_existing

    def all_keys_are_tuples(self) -> bool:
        """Check if all keys are tuples."""
        return all([type(x) == tuple for x in self._data.keys()])

    def get_header(self) -> Dict:
        """Get the header information of the event."""
        return dict(
            name=self.name,
            size=self.size,
            fname=self.fname,
            color=self.color,
            pick_action=self.pick_action,
            win_remove=self.win_remove,
            win_add=self.win_add,
            plot_kwargs=self.plot_kwargs,
            all_keys_are_tuples=self.all_keys_are_tuples(),
        )

    @staticmethod
    def _read_json_file(fname: str) -> Tuple[Dict, Dict]:
        """Read a JSON file and return the header and data."""
        with open(fname, "r") as f:
            header, data = json.load(f)
        if header["all_keys_are_tuples"]:
            data = {eval(k): EventData(**v) for k, v in data.items()}
        else:
            data = {k: EventData(**v) for k, v in data.items()}
        return header, data

    def load(self) -> Tuple[Dict, Dict]:
        """Load the event data from the file."""
        if os.path.exists(self.fname):
            header, data = self._read_json_file(self.fname)
            return header, data
        return {}, {}

    def save(self) -> None:
        """Save the event data to the file."""
        action_str = "Updated" if os.path.exists(self.fname) else "Created"
        with open(self.fname, "w") as f:
            header = self.get_header()
            if header["all_keys_are_tuples"]:
                data = {str(k): v.asdict() for k, v in self._data.items()}
            else:
                data = {k: v.asdict() for k, v in self._data.items()}
            json.dump((header, data), f, indent=4)
        print(action_str + " " + self.fname)

    def add(self, event: Any) -> None:
        """
        Pick the time points of an interval and associate it with a supplied ID.
        If the first selection is outside the axis, then select the first available time point.
        If the last selection is outside the axis, then select the last available time point.
        If the selections are not monotonically increasing, then empty the buffer.
        If any of the 'middle' picks (i.e. not first or last in the sequence) are outside the axes, then empty the buffer.

        The parent UI would invoke this method.

        Args:
            event (Any): The event to add.
        """

        def strictly_increasing(_list: List) -> bool:
            return all(x < y for x, y in zip(_list, _list[1:]))

        def _get_lines() -> List:
            """Return non-empty lines in the axis where event was invoked, or else in all lines in the figure."""
            if event.inaxes is not None:
                return [
                    line
                    for line in event.inaxes.get_lines()
                    if len(line.get_xdata()) > 0
                ]
            return [
                line
                for ax in event.canvas.figure.axes
                for line in ax.get_lines()
                if len(line.get_xdata()) > 0
            ]

        def _get_first_available_timestamp() -> float:
            return min(
                [
                    np.nanmin(l.get_xdata())
                    for l in _get_lines()
                    if len(l.get_xdata()) > 0
                ]
            )

        def _get_last_available_timestamp() -> float:
            return max(
                [
                    np.nanmax(l.get_xdata())
                    for l in _get_lines()
                    if len(l.get_xdata()) > 0
                ]
            )

        def clamp(n: float) -> float:
            smallest = _get_first_available_timestamp()
            largest = _get_last_available_timestamp()
            return max(smallest, min(n, largest))

        if event.xdata is None:
            # click is outside the axis limits
            if not self._buffer:
                inferred_timestamp = _get_first_available_timestamp()
            else:
                assert len(self._buffer) == self.size - 1
                inferred_timestamp = _get_last_available_timestamp()
        else:
            inferred_timestamp = clamp(self.data_func(event.xdata))

        self._buffer.append(inferred_timestamp)

        if not strictly_increasing(self._buffer):
            self._buffer = []

        if len(self._buffer) < self.size:
            return

        assert len(self._buffer) == self.size

        sequence = self._buffer.copy()

        data_id = self.data_id_func()
        if data_id not in self._data:
            self._data[data_id] = EventData()
        if self.pick_action == "append":
            self._data[data_id].added.append(sequence)
        else:  # overwrite
            if len(self._data[data_id].default) not in (0, 1):
                # reset the buffer
                self._buffer = []
                raise AssertionError(
                    "Overwrite mode can only be used when there is at most one default event."
                )
            if len(self._data[data_id].default) == 1:
                # remove the event in default and add the new one to added
                self._data[data_id].removed.append(self._data[data_id].default[0])
                self._data[data_id].default = []
            self._data[data_id].added = [sequence]

        print(self.name, "add", data_id, sequence)
        self._buffer = []
        self.update_display()

    def remove(self, event: Any) -> None:
        """
        Remove an event.

        Args:
            event (Any): The event to remove.
        """
        if event.xdata is None:
            return
        t_marked = float(event.xdata)
        data_id = self.data_id_func()
        if data_id not in self._data:
            return
        ev = self._data[data_id]

        added_start_times = [x[0] for x in ev.added]
        default_start_times = [x[0] for x in ev.default]
        sequence = None
        _removed = False
        _deleted = False
        if len(ev.added) > 0 and len(ev.default) > 0:
            idx_add, val_add = _find_nearest_idx_val(added_start_times, t_marked)
            idx_def, val_def = _find_nearest_idx_val(default_start_times, t_marked)

            add_dist = np.abs(val_add - t_marked)
            def_dist = np.abs(val_def - t_marked)
            if (add_dist <= def_dist) and (
                self.win_remove[0] < add_dist < self.win_remove[1]
            ):
                sequence = ev.added.pop(idx_add)
                _deleted = True
            if (def_dist < add_dist) and (
                self.win_remove[0] < def_dist < self.win_remove[1]
            ):
                sequence = ev.default.pop(idx_def)
                ev.removed.append(sequence)
                _removed = True
        elif len(ev.added) > 0 and len(ev.default) == 0:
            idx_add, val_add = _find_nearest_idx_val(added_start_times, t_marked)
            add_dist = np.abs(val_add - t_marked)
            if self.win_remove[0] < add_dist < self.win_remove[1]:
                sequence = ev.added.pop(idx_add)
                _deleted = True
        elif len(ev.added) == 0 and len(ev.default) > 0:
            idx_def, val_def = _find_nearest_idx_val(default_start_times, t_marked)
            def_dist = np.abs(val_def - t_marked)
            if self.win_remove[0] < def_dist < self.win_remove[1]:
                sequence = ev.default.pop(idx_def)
                ev.removed.append(sequence)
                _removed = True
        else:  # both are empty
            return

        assert sequence is not None

        assert _removed is not _deleted
        # removed moves data from default (i.e. auto-detected) to removed, and delete expunges a manually added event
        print(self.name, {True: "remove", False: "delete"}[_removed], data_id, sequence)
        self.update_display()

    def get_current_event_times(self) -> List:
        """Get the current event times."""
        return self._data.get(self.data_id_func(), EventData()).get_times()

    def _get_display_funcs(self) -> Tuple[Callable, Callable]:
        """Get the display functions."""
        display_type = self.plot_kwargs.get("display_type", "line")
        assert display_type in ("line", "fill")
        if display_type == "fill":
            assert self.size == 2
        if display_type == "line":
            return self._setup_display_line, self._update_display_line
        return self._setup_display_fill, self._update_display_fill

    def setup_display(self) -> None:
        """Setup event display on one or more axes."""
        setup_func, _ = self._get_display_funcs()
        setup_func()

    def _setup_display_line(self) -> None:
        """Setup line display for the event."""
        plot_kwargs = {**{"label": f"event:{self.name}"}, **self.plot_kwargs}
        plot_kwargs.pop("display_type", None)
        for ax in self.ax_list:
            (this_plot,) = ax.plot([], [], color=self.color, **plot_kwargs)
            self.plot_handles.append(this_plot)

    def _setup_display_fill(self) -> None:
        """Setup fill display for the event. Everything is redrawm currently for fill display. So, don't do setup."""
        return

    def update_display(self, draw: bool = True) -> None:
        """Update the event display."""
        _, update_func = self._get_display_funcs()
        update_func(draw)

    def _get_ylim(self, this_ax: plt.Axes, type: str = "data") -> Tuple[float, float]:
        """Get the y-axis limits."""
        if type == "data":
            try:
                def nanlim(x: np.ndarray, default) -> Tuple[float, float]:
                    if np.isnan(x).all():
                        return default
                    return np.nanmin(x), np.nanmax(x)
                x = np.asarray(
                    [
                        nanlim(line.get_xdata(), line.axes.get_ylim())
                        for line in this_ax.get_lines()
                        if not line.get_label().startswith("event:")
                    ]
                )
                return np.nanmin(x[:, 0]), np.nanmax(x[:, 1])
            except ValueError:
                return this_ax.get_ylim()
        return this_ax.get_ylim()

    def _update_display_line(self, draw: bool) -> None:
        """Update the display for 'line' type events."""
        for ax, plot_handle in zip(self.ax_list, self.plot_handles):
            yl = self._get_ylim(ax)
            plot_handle.set_data(
                *utils.ticks_from_times(
                    list(np.asarray(self.get_current_event_times()).flatten()), yl
                )
            )
        if draw:
            plt.draw()

    def _update_display_fill(self, draw: bool) -> None:
        """Update the display for 'fill' type events."""
        if self._hide:
            return
        for plot_handle in self.plot_handles:
            plot_handle.remove()
        self.plot_handles = []
        plot_kwargs = {**dict(alpha=0.2, edgecolor=None), **self.plot_kwargs}
        plot_kwargs.pop("display_type", None)
        for ax in self.ax_list:
            yl = self._get_ylim(ax)
            x = np.asarray(
                [
                    this_times + [np.nan]
                    for this_times in self._data.get(
                        self.data_id_func(), EventData()
                    ).get_times()
                ]
            ).flatten()
            y1 = np.asarray(
                [[yl[0]] * 2 + [np.nan] for _ in range(int(len(x) / 3))]
            ).flatten()
            y2 = np.asarray(
                [[yl[1]] * 2 + [np.nan] for _ in range(int(len(x) / 3))]
            ).flatten()
            this_collection = ax.fill_between(
                x, y1, y2, color=self.color, **plot_kwargs
            )
            self.plot_handles.append(this_collection)
        if draw:
            plt.draw()

    def to_dict(self) -> Dict:
        """Convert the event data to a dictionary."""
        event_data = self._data
        if self.pick_action == "overwrite":
            # keep only the last one - this won't work because get_times will sort the events
            ret = {k: (v.default + v.added)[-1:] for k, v in event_data.items()}
        else:
            ret = {k: v.get_times() for k, v in event_data.items()}
        return ret

    def to_portions(self) -> Dict:
        """Convert the event data to portions."""
        assert self.size == 2
        return {
            signal_id: event_data.to_portions()
            for signal_id, event_data in self._data.items()
        }
        ret = {}
        for signal_id, signal_events in self.to_dict().items():
            ret[signal_id] = functools.reduce(
                lambda a, b: a | b,
                [portion.closed(*interval_limits) for interval_limits in signal_events],
                portion.empty(),
            )
        return ret


class Events(AssetContainer):
    """
    Manager for :py:class:`Event` objects.
    """

    def __init__(self, parent: Any) -> None:
        super().__init__(parent)
        self._text = None

    def add(
        self,
        name: str,
        size: int,
        fname: str,
        data_id_func: Callable,
        color: Union[str, int],
        pick_action: str = "overwrite",
        ax_list: Optional[List] = None,
        win_remove: Tuple[float, float] = (-0.1, 0.1),
        win_add: Tuple[float, float] = (-0.25, 0.25),
        add_key: Optional[str] = None,
        remove_key: Optional[str] = None,
        save_key: Optional[str] = None,
        show: bool = True,
        data_func: Callable = float,
        **plot_kwargs,
    ) -> Event:
        """
        Add an event.

        Args:
            name (str): Name of the event.
            size (int): Length of the sequence. For example, if this is 2, then the event is a pair of start and end times.
            fname (str): File name to save the event.
            data_id_func (Callable): Function to get the current data ID from the parent UI.
            color (Union[str, int]): Color used to display the event
            pick_action (str, optional): Action to take when picking an event ('overwrite' or 'append'). Use overwrite if there can only be one sequence per 'signal'. If there can be multiple, use 'append'. Defaults to overwrite. Defaults to 'overwrite'.
            ax_list (Optional[List], optional): List of axes on which to show the event. Defaults to None.
            win_remove (Tuple[float, float], optional): Window relative to the mouse position to search for removing an event. Defaults to (-0.1, 0.1).
            win_add (Tuple[float, float], optional): Window relative to the mouse position to search for adding an event. Defaults to (-0.25, 0.25). For future use.
            add_key (Optional[str], optional): Keyboard shortcut for adding the event. Defaults to None.
            remove_key (Optional[str], optional): Keyboard shortcut for removing the event. Defaults to None.
            save_key (Optional[str], optional): Keyboard shortcut for saving the events to a JSON file. Defaults to None.
            show (bool, optional): Event visibility. Defaults to True.
            data_func (Callable, optional): Function to apply on the incoming event. Intended use is for type casting. Defaults to float.

        Returns:
            Event: _description_
        """
        assert name not in self.names
        this_ev = Event(
            name,
            size,
            fname,
            data_id_func,
            color,
            pick_action,
            ax_list,
            win_remove,
            win_add,
            data_func,
            **plot_kwargs,
        )
        super().add(this_ev)
        if add_key is not None:
            self.parent.add_key_binding(add_key, this_ev.add, f"Add {name}")
        if remove_key is not None:
            self.parent.add_key_binding(remove_key, this_ev.remove, f"Remove {name}")
        if save_key is not None:
            self.parent.add_key_binding(save_key, this_ev.save, f"Save {name}")
        if show:
            this_ev.setup_display()
        else:
            this_ev._hide = True  # This is for fill displays
        return this_ev

    def add_from_file(
        self,
        fname: str,
        data_id_func: Callable,
        ax_list: Optional[List] = None,
        add_key: Optional[str] = None,
        remove_key: Optional[str] = None,
        save_key: Optional[str] = None,
        show: bool = True,
        data_func: Callable = float,
        **plot_kwargs,
    ) -> Event:
        """Add events from an existing file.
        Intended use case - events are created by another algorithm  meant to be edited using a browser in the :py:mod:`datanavigator`.

        Args:
            fname (str): File name to load the events.
            data_id_func (Callable): Function to get the current data ID from the parent UI.
            ax_list (Optional[List], optional): List of axes on which to show the event. Defaults to None.
            add_key (Optional[str], optional): Keyboard shortcut for adding the event. Defaults to None.
            remove_key (Optional[str], optional): Keyboard shortcut for removing the event. Defaults to None.
            save_key (Optional[str], optional): Keyboard shortcut for saving the events to a JSON file. Defaults to None.
            show (bool, optional): Event visibility. Defaults to True.
            data_func (Callable, optional): Function to apply on the incoming event. Intended use is for type casting. Defaults to float.

        Returns:
            Event: The created Event object.
        """
        assert os.path.exists(fname)
        ev = Event._from_existing_file(fname)
        hdr = ev.get_header()
        del hdr["all_keys_are_tuples"]
        plot_kwargs = {**hdr["plot_kwargs"], **plot_kwargs}
        del hdr["plot_kwargs"]
        return self.add(
            data_id_func=data_id_func,
            ax_list=ax_list,
            add_key=add_key,
            remove_key=remove_key,
            save_key=save_key,
            show=show,
            data_func=data_func,
            **{**hdr, **plot_kwargs},
        )

    def setup_display(self) -> None:
        """Setup display for all events."""
        for ev in self._list:
            ev.setup_display()

    def update_display(self, draw: bool = True) -> None:
        """Update display for all events"""
        for ev in self._list:
            ev.update_display(draw=False)
        if draw:
            plt.draw()


def _find_nearest_idx_val(array: List[float], value: float) -> Tuple[int, float]:
    """Find the index and value of the nearest element in the array to the given value.

    Args:
        array (list[float]): List of values.
        value (float): Value to find the nearest element for.

    Returns:
        Tuple[int, float]: Index and value of the nearest element.
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    val = array[idx]
    return idx, val



"""
Module for browsing videos and extracting clips from videos.

Classes:
    VideoBrowser: Scroll through the frames of a video, extract clips of interest.
    VideoPlotBrowser: Browse a video and an array of `pysampled.Data` side by side.
"""

from __future__ import annotations

import os
from datetime import datetime, timedelta
from typing import Callable, Dict, Optional, Union

import pysampled
from decord import VideoReader
from matplotlib import pyplot as plt
from matplotlib.gridspec import GridSpec

from . import _config
from .core import GenericBrowser


class VideoBrowser(GenericBrowser):
    """Scroll through the frames of a video, extract clips of interest.

    If figure_handle is an axis handle, the video will be plotted in that axis.

    Future Enhancements:
        - Extend VideoBrowser to play, pause, and extract clips using hotkeys.
        - Show timeline in VideoBrowser.
        - Add clickable navigation.
    """

    def __init__(
        self,
        vid_name: str,
        titlefunc: Optional[Callable] = None,
        figure_or_ax_handle: Optional[Union[plt.Axes, plt.Figure]] = None,
        image_process_func: Callable = lambda im: im,
    ):
        """
        Args:
            vid_name (str): Path to the video file.
            titlefunc (Optional[Callable]): Function to generate the title for the plot.
            figure_or_ax_handle (Optional[Union[plt.Axes, plt.Figure]]): Handle to the figure or axis.
            image_process_func (Callable): Function to process the image.
        """
        assert isinstance(figure_or_ax_handle, (plt.Axes, plt.Figure, type(None)))
        if isinstance(figure_or_ax_handle, plt.Axes):
            figure_handle = figure_or_ax_handle.figure
            ax_handle = figure_or_ax_handle
        else:
            # same behavior for figure_or_ax_handle in (None, isinstance(plt.Figure))
            figure_handle = figure_or_ax_handle
            ax_handle = None
        super().__init__(figure_handle)

        if not os.path.exists(vid_name):
            # try looking in the CLIP FOLDER
            vid_name = os.path.join(
                _config.get_clip_folder(), os.path.split(vid_name)[-1]
            )
        assert os.path.exists(vid_name)
        self.fname = vid_name
        self.name = os.path.splitext(os.path.split(vid_name)[1])[0]
        with open(vid_name, "rb") as f:
            self.data = VideoReader(f)

        if ax_handle is None:
            self._ax = self.figure.subplots(1, 1)
        else:
            assert isinstance(ax_handle, plt.Axes)
            self._ax = ax_handle
        self._im = self._ax.imshow(self.data[0].asnumpy())
        self._ax.axis("off")

        self.fps = self.data.get_avg_fps()
        if titlefunc is None:
            self.titlefunc = (
                lambda s: f"Frame {s._current_idx}/{len(s)}, {s.fps} fps, {str(timedelta(seconds=s._current_idx/s.fps))}"
            )

        self.image_process_func = image_process_func

        self.set_default_keybindings()
        self.add_key_binding("e", self.extract_clip)
        self.memoryslots.show(pos="bottom left")

        if self.__class__.__name__ == "VideoBrowser":
            plt.show(block=False)
            self.update()

    def increment_frac(self, n_steps: int = 100) -> None:
        """Browse entire dataset in n_steps.

        Args:
            n_steps (int): Number of steps to increment.
        """
        self._current_idx = min(
            self._current_idx + int(len(self) / n_steps), len(self) - 1
        )
        self.update()

    def decrement_frac(self, n_steps: int = 100) -> None:
        """Browse entire dataset in n_steps.

        Args:
            n_steps (int): Number of steps to decrement.
        """
        self._current_idx = max(self._current_idx - int(len(self) / n_steps), 0)
        self.update()

    def update(self) -> None:
        """Update the video frame."""
        self._im.set_data(
            self.image_process_func(self.data[self._current_idx].asnumpy())
        )
        self._ax.set_title(self.titlefunc(self))
        super().update()
        plt.draw()

    def extract_clip(
        self,
        start_frame: Optional[int] = None,
        end_frame: Optional[int] = None,
        fname_out: Optional[str] = None,
        out_rate: Optional[int] = None,
    ) -> str:
        """Extract a clip from the video.

        Args:
            start_frame (Optional[int]): Starting frame of the clip.
            end_frame (Optional[int]): Ending frame of the clip.
            fname_out (Optional[str]): Output filename.
            out_rate (Optional[int]): Output frame rate.

        Returns:
            str: Path to the extracted clip.
        """
        try:
            import ffmpeg

            use_subprocess = False
        except ModuleNotFoundError:
            import subprocess

            use_subprocess = True

        if start_frame is None:
            start_frame = self.memoryslots._list["1"]
        if end_frame is None:
            end_frame = self.memoryslots._list["2"]
        assert end_frame > start_frame
        start_time = float(start_frame) / self.fps
        end_time = float(end_frame) / self.fps
        dur = end_time - start_time
        if out_rate is None:
            out_rate = self.fps
        if fname_out is None:
            fname_out = os.path.join(
                _config.get_clip_folder(),
                os.path.splitext(self.name)[0]
                + "_s{:.3f}_e{:.3f}.mp4".format(start_time, end_time),
            )
        if use_subprocess:
            subprocess.getoutput(
                f'ffmpeg -ss {start_time} -i "{self.fname}" -r {out_rate} -t {dur} -vcodec h264_nvenc "{fname_out}"'
            )
        else:
            ffmpeg.input(self.fname, ss=start_time).output(
                fname_out, vcodec="h264_nvenc", t=dur, r=out_rate
            ).run()
        return fname_out


class VideoPlotBrowser(GenericBrowser):
    """
    Browse a video and an array of pysampled.Data side by side.

    Args:
        vid_name (str): Path to the video file.
        signals (Dict[str, pysampled.Data]): Dictionary of signals.
        titlefunc (Optional[Callable]): Function to generate the title for the plot.
        figure_handle (Optional[plt.Figure]): Handle to the figure.
        event_win (Optional[tuple]): Event window. Visualize signals around an event. Use this to create "scrolling" plot visualizations, e.g. [-0.5, 1.].
    """

    def __init__(
        self,
        vid_name: str,
        signals: Dict[str, pysampled.Data],
        titlefunc: Optional[Callable] = None,
        figure_handle: Optional[plt.Figure] = None,
        event_win: Optional[tuple] = None,
    ):
        figure_handle = plt.figure(figsize=(20, 12))
        super().__init__(figure_handle)

        self.event_win = event_win

        self.vid_name = vid_name
        assert os.path.exists(vid_name)
        self.name = os.path.splitext(os.path.split(vid_name)[1])[0]
        with open(vid_name, "rb") as f:
            self.video_data = VideoReader(f)
        self.fps = self.video_data.get_avg_fps()

        self.signals = signals
        if titlefunc is None:
            self.titlefunc = (
                lambda s: f"Frame {s._current_idx}/{len(s)}, {s.fps} fps, {str(timedelta(seconds=s._current_idx/s.fps))}"
            )

        self.plot_handles = self._setup()
        self.plot_handles["ax"]["montage"].set_axis_off()

        self.set_default_keybindings()
        self.add_key_binding("e", self.extract_clip)
        self._len = len(self.video_data)
        self.memoryslots.show(pos="bottom left")

        self.figure.canvas.mpl_connect("button_press_event", self.onclick)

        plt.show(block=False)
        self.update()

    def __len__(self) -> int:
        """Return the number of frames in the video."""
        return self._len

    def _setup(self) -> dict:
        """Setup the plot handles.

        Returns:
            dict: Dictionary of plot handles.
        """
        fig = self.figure
        gs = GridSpec(nrows=len(self.signals), ncols=2, width_ratios=[2, 3])
        ax = {}
        plot_handles = {}
        for signal_count, (signal_name, this_signal) in enumerate(self.signals.items()):
            this_ax = fig.add_subplot(gs[signal_count, 1])
            plot_handles[f"signal{signal_count}"] = this_ax.plot(
                this_signal.t, this_signal()
            )
            ylim = this_ax.get_ylim()
            (plot_handles[f"signal{signal_count}_tick"],) = this_ax.plot(
                [0, 0], ylim, "k"
            )
            this_ax.set_title(signal_name)
            if signal_count < len(self.signals) - 1:
                this_ax.get_xaxis().set_ticks([])
            else:
                this_ax.set_xlabel("Time (s)")
            ax[f"signal{signal_count}"] = this_ax

        ax["montage"] = fig.add_subplot(gs[:, 0])
        plot_handles["montage"] = ax["montage"].imshow(self.video_data[0].asnumpy())
        plot_handles["ax"] = ax
        plot_handles["fig"] = fig
        signal_ax = [v for k, v in plot_handles["ax"].items() if "signal" in k]
        signal_ax[0].get_shared_x_axes().join(*signal_ax)
        plot_handles["signal_ax"] = signal_ax
        return plot_handles

    def update(self) -> None:
        """Update the video frame and signals."""
        self.plot_handles["montage"].set_data(
            self.video_data[self._current_idx].asnumpy()
        )
        self.plot_handles["ax"]["montage"].set_title(self.titlefunc(self))
        for signal_count, this_signal in enumerate(self.signals.items()):
            self.plot_handles[f"signal{signal_count}_tick"].set_xdata(
                [self._current_idx / self.fps] * 2
            )
        if self.event_win is not None:
            curr_t = self._current_idx / self.fps
            self.plot_handles["signal_ax"][0].set_xlim(
                curr_t + self.event_win[0], curr_t + self.event_win[1]
            )
        super().update()
        plt.draw()

    def onclick(self, event) -> None:
        """Right click mouse to seek to that frame.

        Args:
            event: Matplotlib event.
        """
        this_frame = round(event.xdata * self.fps)

        if (
            isinstance(this_frame, (int, float))
            and (0 <= this_frame < self._len)
            and (str(event.button) == "MouseButton.RIGHT")
        ):
            self._current_idx = this_frame
            self.update()

    def extract_clip(
        self,
        start_frame: Optional[int] = None,
        end_frame: Optional[int] = None,
        sav_dir: Optional[str] = None,
        out_rate: float = None,
    ) -> str:
        """Save a video of screengrabs.

        Args:
            start_frame (Optional[int]): Starting frame of the clip. Defaults to the first memory slot.
            end_frame (Optional[int]): Ending frame of the clip. Defaults to the second memory slot.
            sav_dir (Optional[str]): Directory to save the clip. If not provided, a timestamped directory is created.

        Returns:
            Path to the saved video file.
        """
        import shutil
        import subprocess

        if start_frame is None:
            start_frame = self.memoryslots._list["1"]
        if end_frame is None:
            end_frame = self.memoryslots._list["2"]
        assert end_frame > start_frame

        if sav_dir is None:
            sav_dir = os.path.join(
                _config.get_clip_folder(), datetime.now().strftime("%Y%m%d_%H%M%S")
            )
        if not os.path.exists(sav_dir):
            os.mkdir(sav_dir)

        if out_rate is None:
            out_rate = self.fps

        print(f"Saving image sequence to {sav_dir}...")
        for frame_count in range(start_frame, end_frame + 1):
            self._current_idx = frame_count
            self.update()
            self.figure.savefig(os.path.join(sav_dir, f"{frame_count:08d}.png"))

        print("Creating video from image sequence...")
        cmd = f'cd "{sav_dir}" && ffmpeg -framerate {self.fps} -start_number {start_frame} -i %08d.png -c:v h264_nvenc -b:v 10M -maxrate 12M -bufsize 24M -vf scale="-1:1080" -an "{sav_dir}.mp4"'
        subprocess.getoutput(cmd)

        print("Removing temporary folder...")
        shutil.rmtree(sav_dir)

        print("Done")
        return f"{sav_dir}.mp4"



import os
import numpy as np
from pathlib import Path
import pytest

import matplotlib.pyplot as plt
from matplotlib import axes as maxes
from datanavigator import get_example_video
from datanavigator.utils import (
    ticks_from_times,
    _List,
    _parse_fax,
    _parse_pos,
    TextView,
    get_palette,
    is_video,
    Video,
    is_pathname_valid,
    is_path_creatable,
    is_path_exists_or_creatable,
)


@pytest.fixture(scope="session", autouse=True)
def initialize_folder_structure(tmp_path_factory):
    base_temp_directory = tmp_path_factory.getbasetemp()

    # Create a sample text file
    (base_temp_directory / "sample_text_file.txt").touch()

    # Download a sammple video file
    video_path = get_example_video(dest_folder=base_temp_directory)
    assert str(video_path).endswith("example_video.mp4")


@pytest.fixture
def video_fname(tmp_path_factory):
    return os.path.join(tmp_path_factory.getbasetemp(), "example_video.mp4")


def test_ticks_from_times():
    times = [1, 2, 3]
    tick_lim = (0, 1)
    x, y = ticks_from_times(times, tick_lim)
    assert x == [1, 1, np.nan, 2, 2, np.nan, 3, 3, np.nan]
    assert y == [0, 1, np.nan, 0, 1, np.nan, 0, 1, np.nan]


def test_List():
    lst = _List([1, 0, 2, 4, 3])
    assert lst.next(1.1) == 2
    assert lst.next(10) == 4
    assert lst.previous(1.1) == 1
    assert lst.previous(1) == lst.previous(0) == lst.previous(-10) == 0


def test_parse_fax():
    fig, ax = plt.subplots()
    f, a = _parse_fax(fig)
    assert isinstance(f, plt.Figure)
    assert isinstance(a, maxes.Axes)
    plt.close(fig)


def test_parse_pos():
    pos = "top left"
    parsed_pos = _parse_pos(pos)
    assert parsed_pos == (0, 1, "top", "left")


def test_TextView():
    text = ["line1", "line2"]
    tv = TextView(text)
    assert tv.text == text
    assert tv._pos == (0, 0, "bottom", "left")
    plt.close(tv.figure)
    tv = TextView({1: "line1", 2: "line2"}, pos="top right")
    assert tv.text == ["1 - line1", "2 - line2"]


def test_get_palette():
    palette = get_palette(n_colors=3)
    assert len(palette) == 3


def test_is_video(tmp_path_factory, video_fname):
    assert is_video(video_fname) == True
    assert is_video(tmp_path_factory.getbasetemp() / "sample_text_file.txt") == False


def test_Video(video_fname):
    video = Video(video_fname)
    assert video.fname == video_fname
    assert video.name == Path(video_fname).stem
    assert len(video) == 300


def test_is_pathname_valid(tmp_path_factory):
    assert is_pathname_valid(str(tmp_path_factory.getbasetemp() / "abc12")) == True
    assert is_pathname_valid(str(tmp_path_factory.getbasetemp() / "<")) == False


def test_is_path_creatable(tmp_path_factory):
    assert is_path_creatable(str(tmp_path_factory.getbasetemp() / "abc12")) == True


def test_is_path_exists_or_creatable(tmp_path_factory):
    assert is_path_exists_or_creatable(str(tmp_path_factory.getbasetemp())) == True
    assert (
        is_path_exists_or_creatable(str(tmp_path_factory.getbasetemp() / "abc12"))
        == True
    )
    assert (
        is_path_exists_or_creatable(str(tmp_path_factory.getbasetemp() / "<")) == False
    )



import pytest
import os
from pathlib import Path
import numpy as np
import pandas as pd
import json
from unittest.mock import patch, MagicMock

import matplotlib.pyplot as plt

import datanavigator
from datanavigator.examples import get_example_video
from tests.conftest import (
    simulate_key_press,
    simulate_key_press_at_xy,
    simulate_mouse_click,
)


@pytest.fixture(scope="module")
def video_fname(tmp_path_factory):
    return get_example_video(dest_folder=str(tmp_path_factory.getbasetemp()))


@pytest.fixture(scope="module")
def ann_fname(video_fname):
    """Fixture to create a temporary JSON file with annotations in the first 10 frames, 3 labels per frame."""
    video = datanavigator.Video(video_fname)
    height, width = video[0].shape[:2]
    ann = datanavigator.VideoAnnotation(vname=video_fname)
    for frame_count in range(10):
        for label in range(3):
            ann.add(
                location=[
                    width * ((frame_count + 10) / 30) + 5 * label,
                    height * ((frame_count + 10) / 30),
                ],
                label=str(label),
                frame_number=frame_count,
            )
    ann.save()
    # load it again to get rid of the empty labels
    ann = datanavigator.VideoAnnotation(fname=ann.fname)
    ann.save()
    return ann.fname


@pytest.fixture(scope="module")
def ann2_fname(video_fname):
    """Fixture to create a second temporary JSON file with 9 annotated frames, 3 labels per frame."""
    video = datanavigator.Video(video_fname)
    height, width = video[0].shape[:2]
    ann2 = datanavigator.VideoAnnotation()
    for frame_count in range(9):
        for label in range(3):
            ann2.add(
                location=[
                    width * ((frame_count - 10) / 30) + 5 * label,
                    height * ((frame_count - 10) / 30),
                ],
                label=str(label),
                frame_number=frame_count * 2 + 50,
            )
    ann2.fname = os.path.join(
        Path(video_fname).parent, Path(video_fname).stem + "_annotations_pn.json"
    )
    ann2.save()
    # loading removes empty labels
    ann2 = datanavigator.VideoAnnotation(fname=ann2.fname)
    ann2.save()
    return ann2.fname


@pytest.fixture(scope="module")
def ann_h5_fname(ann_fname):
    """Fixture to create a temporary HDF5 file of ann."""
    ann = datanavigator.VideoAnnotation(fname=ann_fname)
    ann.to_dlc(save=True)
    return str(Path(ann.fname).with_suffix(".h5"))


@pytest.fixture(scope="function")
def ann_object(ann_fname):
    """Fixture returning a VideoAnnotation object."""
    return datanavigator.VideoAnnotation(fname=ann_fname)


@pytest.fixture(scope="module")
def ann_object_no_video(ann_fname):
    """Fixture returning a VideoAnnotation object without an associated video."""
    # Create a dummy json that doesn't match video name pattern
    data = datanavigator.VideoAnnotation._load_json(ann_fname)
    dummy_fname = Path(ann_fname).parent / "dummy_no_video_annotations.json"
    with open(dummy_fname, "w") as f:
        json.dump(data, f)
    ann = datanavigator.VideoAnnotation(fname=str(dummy_fname))
    yield ann
    os.remove(dummy_fname)  # Clean up dummy file


@pytest.fixture(scope="function")
def ann_object_overlapping(video_fname):
    """Fixture with annotations where all labels overlap for some frames."""
    video = datanavigator.Video(video_fname)
    height, width = video[0].shape[:2]
    ann = datanavigator.VideoAnnotation(vname=video_fname, name="overlapping")
    # Frames 5, 6, 7 have all 3 labels
    for frame_count in range(5, 8):
        for label in range(3):
            ann.add(
                location=[
                    width * ((frame_count + 10) / 30) + 5 * label,
                    height * ((frame_count + 10) / 30),
                ],
                label=str(label),
                frame_number=frame_count,
            )
    # Frame 8 has only label 0
    ann.add(location=[10, 10], label="0", frame_number=8)
    # Frame 9 has labels 0 and 1
    ann.add(location=[10, 10], label="0", frame_number=9)
    ann.add(location=[20, 20], label="1", frame_number=9)
    ann.save()
    # load it again to get rid of the empty labels and ensure correct state
    ann = datanavigator.VideoAnnotation(fname=ann.fname)
    yield ann
    os.remove(ann.fname)  # Clean up


def test_video_annotation_init_empty(tmp_path):
    # test empty initialization
    ann = datanavigator.VideoAnnotation()
    assert ann.fname is None
    assert ann.fstem is None
    assert ann.name == "noname"
    assert ann.video is None
    assert len(ann.data) == len(ann.palette)
    assert all(
        x in ann.plot_handles
        for x in ["ax_list_scatter", "ax_list_trace_x", "ax_list_trace_y"]
    )
    with pytest.raises(AssertionError):
        ann.save()


def test_video_annotation_init_with_video(video_fname):
    video_folder = str(Path(video_fname).parent)
    vstem = Path(video_fname).stem

    # vname only
    ann = datanavigator.VideoAnnotation(vname=video_fname)
    assert ann.fname == os.path.join(video_folder, f"{vstem}_annotations.json")
    assert ann.fstem == f"{vstem}_annotations"
    assert ann.name == "noname"  # default name
    assert ann.video.fname == video_fname


def test_video_annotation_init_with_video_2(video_fname):
    """Supply fname and vname, and omit name"""
    video_folder = str(Path(video_fname).parent)
    vstem = Path(video_fname).stem
    ann_fname = os.path.join(video_folder, f"{vstem}_annotations_pn.json")

    # fname and vname
    ann = datanavigator.VideoAnnotation(fname=ann_fname, vname=video_fname)
    assert ann.fname == os.path.join(video_folder, f"{vstem}_annotations_pn.json")
    assert ann.fstem == f"{vstem}_annotations_pn"
    assert ann.name == "pn"
    assert ann.video.fname == video_fname

    # fname only - find the video
    ann = datanavigator.VideoAnnotation(fname=ann_fname)
    assert ann.fname == os.path.join(video_folder, f"{vstem}_annotations_pn.json")
    assert ann.fstem == f"{vstem}_annotations_pn"
    assert ann.name == "pn"
    assert ann.video.fname == video_fname

    # fname only without "_annotations" in it - this should not be allowed in the future
    ann_fname = os.path.join(video_folder, f"{vstem}_myann.json")
    ann = datanavigator.VideoAnnotation(fname=ann_fname)
    assert ann.fname == ann_fname
    assert ann.fstem == Path(ann_fname).stem
    assert ann.name == "noname"
    assert ann.video is None


def test_video_annotation_init_with_video_3(video_fname):
    """Supply vname and name, and omit fname"""
    video_folder = str(Path(video_fname).parent)
    vstem = Path(video_fname).stem

    # vname and name
    ann = datanavigator.VideoAnnotation(vname=video_fname, name="pn")
    assert ann.fname == os.path.join(video_folder, f"{vstem}_annotations_pn.json")
    assert ann.fstem == f"{vstem}_annotations_pn"
    assert ann.name == "pn"
    assert ann.video.fname == video_fname


def test_video_annotation_load(ann_fname, ann2_fname, ann_h5_fname):
    ann2 = datanavigator.VideoAnnotation(fname=ann2_fname)
    assert ann2.fname == ann2_fname
    assert ann2.fstem == Path(ann2_fname).stem
    assert ann2.name == "pn"
    assert len(ann2.data) == 3
    assert len(ann2.data["0"]) == 9

    # test loading from h5 file
    ann = datanavigator.VideoAnnotation(fname=ann_fname)
    ann_h5 = datanavigator.VideoAnnotation(fname=ann_h5_fname)
    assert ann_h5.fname == ann_h5_fname
    assert ann_h5.fstem == Path(ann_h5_fname).stem
    assert ann_h5.name == "noname"
    assert all([np.allclose(ann.data["0"][i], ann_h5.data["0"][i]) for i in range(10)])


def test_video_annotation_from_multiple_files(
    video_fname, ann_fname, ann2_fname, ann_h5_fname
):
    """Test loading annotations from multiple files."""
    fname_merged = os.path.join(
        Path(video_fname).parent, Path(video_fname).stem + "_annotations_merged.json"
    )
    ann = datanavigator.VideoAnnotation.from_multiple_files(
        fname_list=[ann_fname, ann2_fname],
        vname=video_fname,
        name="merged",
        fname_merged=fname_merged,
    )
    assert ann.fname == fname_merged
    assert len(ann.data) == 3
    assert len(ann.data["0"]) == 19  # 10 + 9 frames

    # test loading from h5 file
    ann_2 = datanavigator.VideoAnnotation.from_multiple_files(
        fname_list=[ann_h5_fname, ann2_fname],
        vname=video_fname,
        name="merged2",
        fname_merged=fname_merged,
    )
    assert ann_2.fname == fname_merged
    assert ann_2.name == "merged2"
    assert len(ann_2.data) == 3
    assert len(ann_2.data["0"]) == 19  # 10 + 9 frames
    assert all([np.allclose(ann.data["0"][i], ann_2.data["0"][i]) for i in range(10)])


def test_video_annotation_properties(
    ann_object, ann_object_no_video, ann_object_overlapping
):
    # n_frames
    assert ann_object.n_frames == len(ann_object.video)
    assert ann_object_no_video.n_frames == 10  # max frame number + 1

    # n_annotations (number of labels)
    assert ann_object.n_annotations == 3
    assert len(ann_object) == 3

    # labels
    assert ann_object.labels == ["0", "1", "2"]

    # frames
    assert ann_object.frames == list(range(10))
    assert ann_object_overlapping.frames == [5, 6, 7, 8, 9]

    # frames_overlapping
    assert ann_object_overlapping.frames_overlapping == [5, 6, 7]
    # Test case with no overlapping frames
    ann_no_overlap = datanavigator.VideoAnnotation()
    ann_no_overlap.data = {}
    ann_no_overlap.add_label("0")
    ann_no_overlap.add_label("1")
    ann_no_overlap.add([1, 1], "0", 0)
    ann_no_overlap.add([1, 1], "1", 1)
    assert ann_no_overlap.frames_overlapping == []


def test_video_annotation_get_frames(ann_object):
    assert ann_object.get_frames("0") == list(range(10))
    assert ann_object.get_frames("1") == list(range(10))
    assert ann_object.get_frames("2") == list(range(10))
    with pytest.raises(AssertionError):
        ann_object.get_frames("nonexistent")


def test_video_annotation_save_errors(ann_object, tmp_path):
    with pytest.raises(ValueError, match="Supply a json file name."):
        ann_object.save(fname=tmp_path / "test.h5")


def test_video_annotation_get_values_cv(ann_object):
    vals = ann_object.get_values_cv(5)
    assert isinstance(vals, np.ndarray)
    assert vals.dtype == np.float32
    assert vals.shape == (3, 1, 2)
    assert not np.isnan(vals).any()

    # Test frame with missing annotations (should not happen in ann_object)
    ann = datanavigator.VideoAnnotation()
    ann.data = {}
    ann.add_label("0")
    ann.add_label("1")
    ann.add([10, 10], "0", 0)
    vals_missing = ann.get_values_cv(0)
    assert vals_missing.shape == (2, 1, 2)
    assert np.isnan(vals_missing[1]).all()


def test_video_annotation_frame_num_str(ann_object, ann_object_no_video):
    assert ann_object._n_digits_in_frame_num() == str(len(str(ann_object.n_frames)))
    assert (
        ann_object._frame_num_as_str(5) == f"{5:0{ann_object._n_digits_in_frame_num()}}"
    )
    assert ann_object_no_video._n_digits_in_frame_num() == str(
        len(str(ann_object_no_video.n_frames))
    )
    assert (
        ann_object_no_video._frame_num_as_str(5)
        == f"{5:0{ann_object_no_video._n_digits_in_frame_num()}}"
    )


def test_video_annotation_add_get_at_frame(ann_object):
    frame_num = 15
    values = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]
    ann_object.add_at_frame(frame_num, values)
    retrieved = ann_object.get_at_frame(frame_num)
    assert retrieved == values
    assert ann_object.data["0"][frame_num] == values[0]
    assert ann_object.data["1"][frame_num] == values[1]
    assert ann_object.data["2"][frame_num] == values[2]

    # Test get_at_frame with missing values
    retrieved_missing = ann_object.get_at_frame(frame_num + 1)
    assert len(retrieved_missing) == 3
    assert all(np.isnan(p).all() for p in retrieved_missing)


def test_video_annotation_getitem(ann_object):
    # Get label data
    assert isinstance(ann_object["0"], dict)
    assert 5 in ann_object["0"]

    # Get frame data
    frame_data = ann_object[5]
    assert isinstance(frame_data, list)
    assert len(frame_data) == 3
    assert frame_data[0] == ann_object.data["0"][5]

    # Get invalid key
    with pytest.raises(ValueError):
        _ = ann_object["invalid_key"]
    with pytest.raises(ValueError):
        _ = ann_object[ann_object.n_frames + 100]  # Non-existent frame


def test_video_annotation_to_dlc_options(ann_object, tmp_path):
    # Test save=False
    df_no_save = ann_object.to_dlc(save=False)
    assert isinstance(df_no_save, pd.DataFrame)
    assert len(df_no_save) == 10  # 10 frames

    # Test custom output path and prefix
    custom_path = tmp_path / "custom_dlc"
    if not os.path.exists(custom_path):
        os.mkdir(custom_path)
    custom_prefix = "my_prefix"
    df_custom = ann_object.to_dlc(
        output_path=custom_path, file_prefix=custom_prefix, save=True
    )
    assert (custom_path / f"{custom_prefix}.h5").exists()
    assert (custom_path / f"{custom_prefix}.csv").exists()
    pd.testing.assert_frame_equal(df_no_save, df_custom)

    # Test dlc prefix
    df_dlc_prefix = ann_object.to_dlc(
        output_path=custom_path, file_prefix="dlc", scorer="tester", save=True
    )
    assert (custom_path / "CollectedData_tester.h5").exists()
    assert (custom_path / "CollectedData_tester.csv").exists()

    # Test internal_to_dlc_labels
    label_map = {"0": "head", "1": "tail", "2": "mid"}
    df_mapped = ann_object.to_dlc(internal_to_dlc_labels=label_map, save=False)
    assert "head" in df_mapped.columns.get_level_values("bodyparts")
    assert "tail" in df_mapped.columns.get_level_values("bodyparts")
    assert "mid" in df_mapped.columns.get_level_values("bodyparts")
    assert "point0" not in df_mapped.columns.get_level_values("bodyparts")


def test_video_annotation_to_traces(ann_object):
    traces = ann_object.to_traces()
    assert isinstance(traces, dict)
    assert set(traces.keys()) == set(ann_object.labels)
    assert traces["0"].shape == (ann_object.n_frames, 2)
    assert traces["1"].shape == (ann_object.n_frames, 2)
    assert traces["2"].shape == (ann_object.n_frames, 2)
    assert not np.isnan(traces["0"][:10]).any()  # First 10 frames are annotated
    assert np.isnan(traces["0"][10:]).all()  # Rest should be NaN


@patch(
    "datanavigator.pointtracking.pysampled", create=True
)  # Mock pysampled if not installed or for isolation
def test_video_annotation_to_signal(mock_pysampled, ann_object, ann_object_no_video):
    mock_data = MagicMock()
    mock_pysampled.Data.return_value = mock_data

    # Test with video
    signal_0 = ann_object.to_signal("0")
    mock_pysampled.Data.assert_called_once()
    call_args, call_kwargs = mock_pysampled.Data.call_args
    np.testing.assert_array_equal(call_args[0], ann_object.to_trace("0"))
    assert call_kwargs["sr"] == ann_object.video.get_avg_fps()
    assert signal_0 == mock_data

    # Test multiple signals
    mock_pysampled.Data.reset_mock()
    signals = ann_object.to_signals()
    assert mock_pysampled.Data.call_count == 3
    assert set(signals.keys()) == set(ann_object.labels)
    assert signals["0"] == mock_data
    assert signals["1"] == mock_data
    assert signals["2"] == mock_data

    # Test without video (should raise error)
    mock_pysampled.Data.reset_mock()
    with pytest.raises(AssertionError):
        ann_object_no_video.to_signal("0")
    with pytest.raises(AssertionError):
        ann_object_no_video.to_signals()


def test_video_annotation_add_label():
    ann = datanavigator.VideoAnnotation()
    assert ann.labels == [str(i) for i in range(10)]  # Starts with 10 empty labels

    # Add data to existing label
    ann.add([1, 1], "0", 0)
    assert ann.data["0"] == {0: [1, 1]}

    # Create a new instance to test adding labels beyond initial empty ones
    ann_new = datanavigator.VideoAnnotation()
    ann_new.data = {}  # Start truly empty for this test

    # Add label automatically
    ann_new.add_label()
    assert ann_new.labels == ["0"]
    assert ann_new.data["0"] == {}
    assert ann_new.palette[list(ann_new.data.keys()).index("0")] is not None  # Color assigned

    # Add specific label
    ann_new.add_label("5")
    assert ann_new.labels == ["0", "5"]
    assert ann_new.data["5"] == {}
    assert ann_new.palette[list(ann_new.data.keys()).index("0")] is not None

    # Add with specific color
    ann_new.add_label("2", color=(0.1, 0.2, 0.3))
    assert ann_new.labels == ["0", "2", "5"]
    assert ann_new.data["2"] == {}

    # Add existing label (no longer fails)
    ann_new.add_label("0")

    # Add invalid label (non-digit string)
    with pytest.raises(AssertionError):
        ann_new.add_label("a")

    # Add more than 10 labels (fill up first)
    for i in [1, 3, 4, 6, 7, 8, 9]:
        ann_new.add_label(str(i))
    assert len(ann_new.labels) == 10
    ann_new.add_label()  # Cannot add 11th label automatically
    assert "10" in ann_new.labels  # Check that it was added as a new label


def test_video_annotation_remove(ann_object):
    assert 5 in ann_object.data["0"]
    ann_object.remove("0", 5)
    assert 5 not in ann_object.data["0"]
    # Remove non-existent frame (should not fail)
    ann_object.remove("0", 500)
    # Remove non-existent label (should fail)
    with pytest.raises(AssertionError):
        ann_object.remove("nonexistent", 0)


def test_video_annotation_clip_labels(ann_object_overlapping):
    # Original frames: [5, 6, 7, 8, 9]
    ann_object_overlapping.clip_labels(start_frame=6, end_frame=8)
    assert ann_object_overlapping.frames == [6, 7, 8]
    assert 5 not in ann_object_overlapping.data["0"]
    assert 9 not in ann_object_overlapping.data["0"]
    assert 6 in ann_object_overlapping.data["0"]
    assert 7 in ann_object_overlapping.data["0"]
    assert 8 in ann_object_overlapping.data["0"]
    assert 6 in ann_object_overlapping.data["1"]  # Check other labels too
    assert 7 in ann_object_overlapping.data["1"]
    assert 8 not in ann_object_overlapping.data["1"]  # Label 1 didn't exist at frame 8


def test_video_annotation_keep_overlapping_continuous_frames(ann_object_overlapping):
    # Original frames: [5, 6, 7, 8, 9]
    # Overlapping frames: [5, 6, 7]
    # Continuous overlapping: [5, 6], [6, 7] -> keep 5, 6, 7
    assert set(ann_object_overlapping.frames) == set([5, 6, 7, 8, 9])
    ann_object_overlapping.keep_overlapping_continuous_frames()
    assert set(ann_object_overlapping.frames) == set([5, 6, 7])
    assert 8 not in ann_object_overlapping.data["0"]
    assert 9 not in ann_object_overlapping.data["0"]
    assert 5 in ann_object_overlapping.data["0"]
    assert 6 in ann_object_overlapping.data["1"]
    assert 7 in ann_object_overlapping.data["2"]

    # Test case with no continuous overlapping frames
    ann = datanavigator.VideoAnnotation()
    ann.data = {}
    ann.add_label("0")
    ann.add_label("1")
    ann.add([1, 1], "0", 0)
    ann.add([1, 1], "1", 0)
    ann.add([1, 1], "0", 2)
    ann.add([1, 1], "1", 2)
    ann.keep_overlapping_continuous_frames()  # Should print warning and do nothing
    assert ann.frames == [0, 2]


@patch("datanavigator.pointtracking.pysampled", create=True)  # Mock pysampled
def test_video_annotation_get_area(mock_pysampled, ann_object, ann_object_no_video):
    mock_data = MagicMock()
    mock_pysampled.Data.return_value = mock_data

    # Test with video -> returns pysampled.Data
    area_signal = ann_object.get_area(labels=["0", "1", "2"])
    assert area_signal == mock_data
    mock_pysampled.Data.assert_called_once()
    call_args, call_kwargs = mock_pysampled.Data.call_args
    assert call_args[0].shape == (ann_object.n_frames,)
    assert call_kwargs["sr"] == ann_object.video.get_avg_fps()

    # Test with string labels
    mock_pysampled.Data.reset_mock()
    area_signal_str = ann_object.get_area(labels="012")
    assert area_signal_str == mock_data
    mock_pysampled.Data.assert_called_once()

    # Test with lowpass
    mock_pysampled.Data.reset_mock()
    mock_signal = MagicMock()
    mock_signal.lowpass.return_value.return_value = ann_object.to_trace(
        "0"
    )  # Mock lowpass output
    with patch.object(
        ann_object,
        "to_signals",
        return_value={"0": mock_signal, "1": mock_signal, "2": mock_signal},
    ):
        area_signal_lp = ann_object.get_area(labels="012", lowpass=5)
        assert area_signal_lp == mock_data
        mock_signal.lowpass.assert_called_with(5)
        mock_pysampled.Data.assert_called_once()

    # Test without video -> returns np.ndarray
    mock_pysampled.Data.reset_mock()
    area_array = ann_object_no_video.get_area(labels=["0", "1", "2"])
    assert isinstance(area_array, np.ndarray)
    assert area_array.shape == (ann_object_no_video.n_frames,)
    mock_pysampled.Data.assert_not_called()

    # Test invalid label
    with pytest.raises(AssertionError):
        ann_object.get_area(labels=["0", "invalid"])


def test_video_annotation_init_preloaded(ann_fname):
    # Load data first
    preloaded_data = datanavigator.VideoAnnotation._load_json(ann_fname)
    # Init with preloaded data
    ann = datanavigator.VideoAnnotation(preloaded_json=preloaded_data)
    assert ann.fname is None  # No fname provided
    assert ann.video is None  # No vname provided
    assert ann.name == "noname"
    assert ann.data == preloaded_data
    assert len(ann.labels) == 3


def test_video_annotation_display_setup(video_fname):
    # Use spec=matplotlib.axes.Axes to make isinstance checks pass
    fig, (ax_img, ax_x, ax_y) = plt.subplots(3, 1)
    ann = datanavigator.VideoAnnotation(vname=video_fname, name="test_display_setup")

    # Setup scatter
    ann.setup_display_scatter(ax_list_scatter=[ax_img])
    assert "labels_in_ax0" in ann.plot_handles

    # Setup trace
    ann.setup_display_trace(ax_list_trace_x=[ax_x], ax_list_trace_y=[ax_y])
    # Called 10 times for x and 10 times for y (for labels 0-9)
    assert "trace_in_axx0_label0" in ann.plot_handles
    assert "trace_in_axy0_label0" in ann.plot_handles

    plt.close(fig)


def test_video_annotation_display_update_visibility(video_fname):
    fig, (ax_img, ax_x, ax_y) = plt.subplots(3, 1)

    ann = datanavigator.VideoAnnotation(
        vname=video_fname, name="test_display_update_visibility"
    )
    ann.setup_display(
        ax_list_scatter=[ax_img], ax_list_trace_x=[ax_x], ax_list_trace_y=[ax_y]
    )
    ann.add([1, 1], "0", 0)  # Add some data

    # Update scatter
    ann.update_display_scatter(0)

    # Update trace
    ann.update_display_trace("0")
    # Called for x and y trace of label 0

    # Update display calls both
    ann.update_display(0, "0")

    # Test visibility/alpha/plot_type
    ann.hide(draw=False)

    ann.show(draw=False)

    ann.set_alpha(0.5, draw=False)

    ann.set_plot_type("line", draw=False)

    ann.set_plot_type("dot", draw=False)

    # Test show/hide one trace
    ann.hide_trace("0", draw=False)

    ann.show_one_trace("1", draw=False)
    plt.close(fig)
    assert True  # If no exceptions, test passed


def test_video_point_annotator_init(video_fname):
    # simple initialization with video only.
    v = datanavigator.VideoPointAnnotator(vid_name=video_fname)
    assert len(v.annotations) == 2
    assert v.annotations[0].name == ""
    assert v.annotations[0].fstem == "example_video_annotations"
    assert v.annotations[-1].name == "buffer"
    plt.close(v.figure)

    v = datanavigator.VideoPointAnnotator(vid_name=video_fname, annotation_names=["pn"])
    assert v.annotations.names == [
        "pn",
        "buffer",
    ]  # the _annotations.json file will not be loaded
    plt.close(v.figure)

    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["", "pn"]
    )
    assert v.annotations.names == ["", "pn", "buffer"]
    plt.close(v.figure)

    # explicitly adding a "buffer" name will not make a difference
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["", "pn", "buffer"]
    )
    assert v.annotations.names == ["", "pn", "buffer"]
    plt.close(v.figure)


def test_video_point_annotator_from_annotations(ann_object):
    v = datanavigator.VideoPointAnnotator.from_annotations(ann_object)
    assert len(v.ann.frames) == 10


def test_video_point_annotator_add_new_label(ann_object):
    v = datanavigator.VideoPointAnnotator.from_annotations(ann_object)
    assert "5" not in v.ann.labels
    assert "5" not in v.annotations["buffer"].labels
    v(simulate_key_press(v.figure, key="5"))
    assert "5" in v.ann.labels
    assert "5" in v.annotations["buffer"].labels
    assert len(v.ann.data["5"]) == 0

    # test "place" mode
    v(simulate_key_press(v.figure, key="`"))  # toggle what number keys do
    assert v.statevariables["number_keys"].current_state == "place"
    assert "6" not in v.ann.labels
    v(simulate_key_press_at_xy((v.figure, v._ax_image), key="6", xdata=645, ydata=360))
    assert "6" in v.ann.labels
    assert len(v.ann.data["6"]) == 1
    assert np.allclose(v.ann["6"][v._current_idx], [645, 360])
    assert (
        len(v.annotations["buffer"].frames) == 0
    )  # buffer should still be empty if we added the annotation to the current layer


def test_video_point_annotator_key_bindings(video_fname):
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["", "pn"]
    )
    assert len(v.annotations[""].frames) == 10
    assert len(v.annotations["pn"].frames) == 9
    assert len(v.annotations["buffer"].frames) == 0

    # check state variables
    assert v.statevariables.names == ['annotation_layer', 'annotation_overlay', 'annotation_label', 'label_range', 'number_keys']

    # check the current state
    assert v._current_idx == 0
    assert v._current_label == v.statevariables["annotation_label"].current_state == "0"
    assert v._current_layer == v.statevariables["annotation_layer"].current_state == ""
    assert (
        v._current_overlay is None
        and v.statevariables["annotation_overlay"].current_state is None
    )

    # cycle through the annotation layers
    assert v.ann.name == ""
    v(simulate_key_press(v.figure, key="="))
    assert v.ann.name == "pn"
    v(simulate_key_press(v.figure, key="="))
    assert v.ann.name == "buffer"
    v(simulate_key_press(v.figure, key="="))
    assert v.ann.name == ""
    v(simulate_key_press(v.figure, key="="))
    assert v.ann.name == "pn"

    # in layer pn, go to frame 18
    v(simulate_key_press(v.figure, key="'"))  # go to annotation label 1
    assert v._current_label == "1"
    event = simulate_mouse_click(
        (v.figure, v._ax_trace_x),
        xdata=18,
        ydata=np.mean(v._ax_trace_x.get_ylim()),
        button=3,  # Right click
    )
    v.figure.canvas.callbacks.process("button_press_event", event)
    assert v._current_idx == 18

    # annotate a point with label "1" at frame 18 in layer "pn"
    assert v._current_idx not in v.ann[v._current_label]
    v(simulate_key_press_at_xy((v.figure, v._ax_image), key="t", xdata=645, ydata=360))
    assert np.allclose(v.ann[v._current_label][v._current_idx], [645, 360])

    # move this point with right mouse click on the image
    event = simulate_mouse_click(
        (v.figure, v._ax_image),
        xdata=700,
        ydata=400,
        button=3,  # Right click
    )
    v.figure.canvas.callbacks.process("button_press_event", event)
    assert np.allclose(v.ann[v._current_label][v._current_idx], [700, 400])

    # check that the pn layer has 1 annotation at frame 18, and 9 existing annotations
    assert v.ann.frames == list(np.r_[18, 50 : 50 + 18 : 2])

    # make buffer the annotation overlay
    v(simulate_key_press(v.figure, key="["))
    assert v._current_overlay == "buffer"

    # interpolate with lk (check labels with lk - puts interpolated points in a buffer layer)
    # test1 - minimal mode - two previous labeled frames to two next frames (or 1 next frame if current frame is annotated)
    assert (
        len(v.annotations["buffer"].frames) == 0
    )  # check that the buffer layer is empty
    v(simulate_key_press(v.figure, key="v"))
    assert v.annotations["buffer"].frames == list(np.r_[18:51])

    v._current_idx = 19
    v.update()
    v(simulate_key_press(v.figure, key="v"))
    assert v.annotations["buffer"].frames == list(np.r_[18:53])

    v._current_idx = 51
    v.update()
    v(simulate_key_press(v.figure, key="v"))
    assert v.annotations["buffer"].frames == list(np.r_[18:55])
    assert len(v.annotations["buffer"].frames) == 37  # 18 to 54, including 18 and 54

    # clear the buffer layer
    # bring the buffer layer to the foreground
    v(simulate_key_press(v.figure, key="="))
    assert v._current_layer == "buffer"
    # select an event using the 'z' key
    for xdata in (4, 85):
        v(
            simulate_key_press_at_xy(
                (v.figure, v._ax_trace_y),
                key="z",
                xdata=xdata,
                ydata=400,
            )
        )
    v(simulate_key_press(v.figure, key="alt+a"))  # clear the buffer layer
    assert (
        len(v.annotations["buffer"].frames) == 0
    )  # check that the buffer layer is empty
    plt.close(v.figure)


def test_video_point_annotator_add_remove_annotation(video_fname):
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn2"]
    )
    assert len(v.annotations) == 2
    assert v.annotations.names == ["pn2", "buffer"]

    assert v.statevariables["number_keys"].current_state == "select"
    v(simulate_key_press(v.figure, key="`"))  # toggle what number keys do
    assert v.statevariables["number_keys"].current_state == "place"

    # place 3 pointso on the current frame (addubg with "t" is already covered in the previous test)
    v._current_idx = 45  # go to frame 45
    v.update()
    assert v._current_layer == "pn2"
    for label in ("0", "1", "2"):
        v(
            simulate_key_press_at_xy(
                (v.figure, v._ax_image),
                key=label,
                xdata=100 + int(label) * 10,
                ydata=200 - int(label) * 10,
            )
        )
    assert v.ann.frames == [45]
    assert np.allclose(v.ann["0"][v._current_idx], [100, 200])
    assert np.allclose(v.ann["1"][v._current_idx], [110, 190])
    assert np.allclose(v.ann["2"][v._current_idx], [120, 180])
    assert v._current_label == "2"  # last label used is the current label

    # remove annotation for point 2
    assert len(v.ann["2"]) == 1  # check that the annotation for point 2 exists
    v(simulate_key_press(v.figure, key="y"))
    assert len(v.ann["2"]) == 0  # check that the annotation for point 2 is removed
    plt.close(v.figure)


def test_video_point_annotator_conditional_move(video_fname):
    """Test the feature that allows to change frame only if there is no annotation at the current frame."""
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn2"]
    )
    assert len(v.annotations) == 2
    assert v.annotations.names == ["pn2", "buffer"]

    # add a point at frame 10
    v._current_idx = 10  # go to frame 0
    v.update()
    v(
        simulate_key_press_at_xy(
            (v.figure, v._ax_image),
            key="t",
            xdata=100,
            ydata=200,
        )
    )
    assert np.allclose(v.ann["0"][v._current_idx], [100, 200])

    v(simulate_key_press(v.figure, key="left"))
    assert v._current_idx == 9  # go to frame 9
    v(simulate_key_press(v.figure, key="f"))
    assert v._current_idx == 10
    v(simulate_key_press(v.figure, key="f"))
    assert (
        v._current_idx == 10
    )  # shouldn not move because there is an annotation at frame 10

    v(simulate_key_press(v.figure, key="g"))
    assert v._current_idx == 11  # go to frame 9
    v(simulate_key_press(v.figure, key="d"))
    assert v._current_idx == 10  # go to frame 10
    v(simulate_key_press(v.figure, key="d"))
    assert (
        v._current_idx == 10
    )  # shouldn't move because there is an annotation at frame 10


def test_video_point_annotator_state_variables(video_fname):
    """Test cycling through the state variables."""
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["", "pn"]
    )
    assert len(v.annotations) == 3
    assert v.annotations.names == ["", "pn", "buffer"]

    # check state variables
    assert v.statevariables.names == ['annotation_layer', 'annotation_overlay', 'annotation_label', 'label_range', 'number_keys']

    # check the current state
    assert v._current_idx == 0
    assert v._current_label == v.statevariables["annotation_label"].current_state == "0"
    assert v._current_layer == v.statevariables["annotation_layer"].current_state == ""
    assert (
        v._current_overlay is None
        and v.statevariables["annotation_overlay"].current_state is None
    )

    assert v._current_layer == ""
    v(simulate_key_press(v.figure, key="="))
    assert v._current_layer == "pn"
    v(simulate_key_press(v.figure, key="-"))
    assert v._current_layer == ""
    v(simulate_key_press(v.figure, key="-"))
    assert v._current_layer == "buffer"
    v(simulate_key_press(v.figure, key="="))
    assert v._current_layer == ""

    assert v._current_overlay is None
    v(simulate_key_press(v.figure, key="["))
    assert v._current_overlay == "buffer"
    v(simulate_key_press(v.figure, key="]"))
    assert v._current_overlay is None
    v(simulate_key_press(v.figure, key="]"))
    assert v._current_overlay == ""
    v(simulate_key_press(v.figure, key="]"))
    assert v._current_overlay == "pn"

    assert v.ann.labels == ["0", "1", "2"]
    assert v._current_label == "0"
    v(simulate_key_press(v.figure, key="'"))
    assert v._current_label == "1"
    v(simulate_key_press(v.figure, key=";"))
    assert v._current_label == "0"
    v(simulate_key_press(v.figure, key=";"))
    assert v._current_label == "2"

    v(simulate_key_press(v.figure, key="="))
    assert v._current_layer == "pn"
    assert v.ann.frames == list(
        np.r_[50 : 50 + 18 : 2]
    )  # check that the annotation layer is empty
    assert v._current_idx == 0
    v(simulate_key_press(v.figure, key="."))
    assert v._current_idx == 50
    v(simulate_key_press(v.figure, key="."))
    assert v._current_idx == 52
    v(simulate_key_press(v.figure, key=","))
    assert v._current_idx == 50
    v(simulate_key_press(v.figure, key=","))
    assert v._current_idx == 50

    plt.close(v.figure)


def test_video_point_annotator_frames_of_interest(video_fname):
    """Test the frames of interest feature."""
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn3"]
    )
    assert len(v.annotations) == 2
    assert v.annotations.names == ["pn3", "buffer"]

    # check initial state
    assert v.frames_of_interest == []

    # add a frame of interest
    assert v._current_layer == "pn3"
    # test navigation with empty frames of interest
    v(simulate_key_press(v.figure, key="alt+,"))  # should not throw an error
    v(simulate_key_press(v.figure, key="alt+."))  # should not throw an error

    v._current_idx = 4
    v.update()
    v(simulate_key_press(v.figure, key="m", inaxes=v._ax_trace_x))
    assert v.frames_of_interest == [4]
    v(simulate_key_press(v.figure, key="m", inaxes=v._ax_image))  # this should not work
    assert v.frames_of_interest == [4]
    v(simulate_key_press(v.figure, key="m"))  # this should also not work
    assert v.frames_of_interest == [4]

    # add and remove a frame of interest
    v._current_idx = 7
    v.update()
    v(simulate_key_press(v.figure, key="m", inaxes=v._ax_trace_x))
    assert v.frames_of_interest == [4, 7]
    v(simulate_key_press(v.figure, key="alt+,"))
    assert v._current_idx == 4
    v(simulate_key_press(v.figure, key="alt+,"))
    assert v._current_idx == 4
    v(simulate_key_press(v.figure, key="alt+."))
    assert v._current_idx == 7
    v(simulate_key_press(v.figure, key="alt+."))
    assert v._current_idx == 7

    v(simulate_key_press(v.figure, key="m", inaxes=v._ax_trace_y))
    assert v.frames_of_interest == [4]

    # label two frames, interpolate between them into the buffer, and copy frames of interest from the buffer
    # label two frames
    event = simulate_mouse_click(
        (v.figure, v._ax_trace_x),
        xdata=27.1,
        ydata=np.mean(v._ax_trace_x.get_ylim()),
        button=3,  # Right click
    )
    v.figure.canvas.callbacks.process("button_press_event", event)
    assert v._current_idx == 27
    assert v._current_idx not in v.ann[v._current_label]
    v(simulate_key_press_at_xy((v.figure, v._ax_image), key="t", xdata=441, ydata=380))
    assert np.allclose(v.ann[v._current_label][v._current_idx], [441, 380])

    event = simulate_mouse_click(
        (v.figure, v._ax_trace_x),
        xdata=60.1,
        ydata=np.mean(v._ax_trace_x.get_ylim()),
        button=3,  # Right click
    )
    v.figure.canvas.callbacks.process("button_press_event", event)
    assert v._current_idx == 60
    assert v._current_idx not in v.ann[v._current_label]
    v(simulate_key_press_at_xy((v.figure, v._ax_image), key="t", xdata=468, ydata=487))
    assert np.allclose(v.ann[v._current_label][v._current_idx], [468, 487])

    # bring buffer to overlay
    v(simulate_key_press(v.figure, key="["))
    assert v._current_overlay == "buffer"

    # interpolate with lk (check labels with lk - puts interpolated points in a buffer layer)
    assert (
        len(v.annotations["buffer"].frames) == 0
    )  # check that the buffer layer is empty
    v(simulate_key_press(v.figure, key="v"))
    assert v.annotations["buffer"].frames == list(np.r_[27:61])

    # mark frames of interest as 31, 42, 52
    v.frames_of_interest = [31, 42, 52]
    v.update()

    # copy frames of interest from the buffer to the annotation layer
    v(simulate_key_press(v.figure, key="alt+c"))

    assert v.ann.frames == [27, 31, 42, 52, 60]

    plt.close(v.figure)


def test_video_point_annotator_copy_annotations(video_fname):
    """Test the feature of copying annotations between layers."""
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn4", "pn5"]
    )
    # testing c - copy current annotaion from overlay to the current annotation layer
    # add a label to pn4
    v.annotations["pn4"].add([1, 1], "1", 5)

    # go to frame 5
    event = simulate_mouse_click(
        (v.figure, v._ax_trace_x),
        xdata=5,
        ydata=np.mean(v._ax_trace_x.get_ylim()),
        button=3,  # Right click
    )
    v.figure.canvas.callbacks.process("button_press_event", event)
    assert v._current_idx == 5

    # set current layer to pn5
    v(simulate_key_press(v.figure, key="="))
    assert v._current_layer == "pn5"

    # set current overlay to pn4
    v(simulate_key_press(v.figure, key="]"))
    assert v._current_overlay == "pn4"

    # check current state of the data
    assert v.ann.frames == []
    assert v.annotations["pn4"].frames == [5]

    # copy annotation from overlay to current layer
    v(
        simulate_key_press(v.figure, key="c")
    )  # nothing should happen because current label is 0
    assert v.ann.frames == []

    # set current label to 1
    v(simulate_key_press(v.figure, key="'"))

    # copy annotation from overlay to current layer
    v(
        simulate_key_press(v.figure, key="c")
    )  # copy annotation from overlay to current layer
    assert v.ann.frames == [5]
    assert len(v.ann.data["1"]) == 1
    assert all([len(v.ann.data[x]) == 0 for x in v.ann.labels if x != "1"])

    # remove annotation from the current layer
    v(simulate_key_press(v.figure, key="y"))  # remove annotation from the current layer
    assert all([len(v.ann.data[x]) == 0 for x in v.ann.labels])

    # testing alt+c - copy frames of interest from overlay to current layer
    # add data into the buffer layer from frames 7 to 50
    for i in range(7, 51):
        v.annotations["buffer"].add([i, i], "1", i)
    assert (
        len(v.annotations["buffer"].frames) == 44
    )  # check that the buffer layer has 44 frames

    # add frames of interest
    v.frames_of_interest = [2, 20, 23]  # there is no data in 2
    v.update()

    # copy frames of interest from buffer to current layer
    assert v._current_layer == "pn5"
    assert list(v.annotations["pn5"].data["1"].keys()) == []
    assert v._current_overlay != "buffer"
    v(
        simulate_key_press(v.figure, key="alt+c")
    )  # copy frames of interest from buffer to current layer
    assert list(v.annotations["pn5"].data["1"].keys()) == []

    # should only work after changing the overlay to buffer
    v(simulate_key_press(v.figure, key="]"))
    v(simulate_key_press(v.figure, key="]"))
    assert v._current_overlay == "buffer"
    v(
        simulate_key_press(v.figure, key="alt+c")
    )  # copy frames of interest from buffer to current layer
    assert list(v.annotations["pn5"].data["1"].keys()) == [20, 23]

    # ctrl+alt+c - copy frames in interval from overlay to current layer.
    # change current layer to pn4 and overlay to pn5
    v(simulate_key_press(v.figure, key="="))
    v(simulate_key_press(v.figure, key="="))
    assert v._current_layer == "pn4"

    # add interval to the current layer - select an event using the 'z' key
    for xdata in (30, 70):
        v(
            simulate_key_press_at_xy(
                (v.figure, v._ax_trace_y),
                key="z",
                xdata=xdata,
                ydata=v._ax_trace_y.get_ylim()[0],
            )
        )
    assert v.annotations[v._current_overlay].frames == list(np.r_[7:51])
    v(
        simulate_key_press(v.figure, key="ctrl+alt+c")
    )  # copy interval from overlay to current layer
    assert v.ann.frames == list(np.r_[5, 30:51])

    # copy all annotation from overlay to current layer in the current frame (no shortcut)
    # go to frame 12
    event = simulate_mouse_click(
        (v.figure, v._ax_trace_x),
        xdata=12,
        ydata=np.mean(v._ax_trace_x.get_ylim()),
        button=3,  # Right click
    )
    v.figure.canvas.callbacks.process("button_press_event", event)
    assert v._current_idx == 12
    assert v._current_layer == "pn4"
    assert v._current_overlay == "buffer"
    assert v._current_label == "1"

    # add data to labels 3 and 4 in the buffer layer
    v.annotations["buffer"].add([100, 13], "3", 12)
    v.annotations["buffer"].add([100, 123], "4", 12)

    assert 12 not in v.ann.frames
    v.copy_annotations_from_overlay()
    assert 12 in v.ann.frames
    assert list(v.ann.data["3"].keys()) == [12]
    assert list(v.ann.data["4"].keys()) == [12]
    assert np.allclose(v.ann["3"][12], [100, 13])
    assert np.allclose(v.ann["4"][12], [100, 123])
    assert all(
        [12 not in v.ann.data[x] for x in v.ann.labels if x not in ["1", "3", "4"]]
    )

    plt.close(v.figure)


def test_video_point_annotator_lucas_kanade(video_fname):
    # -- predict_points_with_lucas_kanade --
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn6"]
    )
    v.ann.data = {}
    v.ann.add_label("0")
    v.ann.add_label("1")
    v.ann.add_label("2")
    v.ann.add_label("3")

    # add some annotations in frame 5
    v.annotations["pn6"].add([10, 10], "0", 5)
    v.annotations["pn6"].add([10, 11], "1", 5)
    v.annotations["pn6"].add([10, 12], "2", 5)

    # go to frame 8
    event = simulate_mouse_click(
        (v.figure, v._ax_trace_x),
        xdata=8,
        ydata=np.mean(v._ax_trace_x.get_ylim()),
        button=3,  # Right click
    )
    v.figure.canvas.callbacks.process("button_press_event", event)
    assert v._current_idx == 8

    # alt+b - predict current point with lucas kanade and add to the current layer
    assert v._current_label == "0"
    assert 8 not in v.annotations["pn6"].frames
    v(simulate_key_press(v.figure, key="alt+b"))
    assert 8 in v.annotations["pn6"]["0"]
    assert 8 not in v.annotations["pn6"]["1"]
    assert 8 not in v.annotations["pn6"]["2"]
    assert 8 not in v.annotations["pn6"]["3"]

    # ctrl+b - predict all points with lucas kanade and add to the current layer
    # the nearest annotated frame for ALL labels should be the same, otherwise this will not work
    with pytest.raises(ValueError):
        # because there is no data for label 3 in frame 5
        v(simulate_key_press(v.figure, key="ctrl+b"))
    v.annotations["pn6"].add([10, 13], "3", 5)

    with pytest.raises(AssertionError):
        # because different labels have different nearest annotated frames
        v(simulate_key_press(v.figure, key="ctrl+b"))
    v(simulate_key_press(v.figure, key="y"))  # otherwise the next test will fail
    v.update()

    v(simulate_key_press(v.figure, key="ctrl+b"))
    assert 8 in v.annotations["pn6"]["1"]
    assert 8 in v.annotations["pn6"]["2"]
    assert 8 in v.annotations["pn6"]["3"]

    # change overlay to pn6 and current layer to buffer
    v(simulate_key_press(v.figure, key="]"))
    assert v._current_overlay == "pn6"
    v(simulate_key_press(v.figure, key="="))
    assert v._current_layer == "buffer"

    # add an interval to the current layer - select an event using the 'z' key
    for xdata in (6, 40):
        v(
            simulate_key_press_at_xy(
                (v.figure, v._ax_trace_y),
                key="z",
                xdata=xdata,
                ydata=v._ax_trace_y.get_ylim()[0],
            )
        )
    with pytest.raises(KeyError):
        # doesn't work if the starting point of the interval is not in the overlay
        v(simulate_key_press(v.figure, key="ctrl+d"))

    for xdata in (8, 40):
        v(
            simulate_key_press_at_xy(
                (v.figure, v._ax_trace_y),
                key="z",
                xdata=xdata,
                ydata=v._ax_trace_y.get_ylim()[0],
            )
        )
    v(simulate_key_press(v.figure, key="ctrl+d"))

    plt.close(v.figure)


def test_video_point_annotator_prev_next_frames_with_current_label(video_fname):
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn7"]
    )
    # add annotation in frames 5 and 8
    v.annotations["pn7"].add([10, 10], "0", 5)
    v.annotations["pn7"].add([10, 11], "0", 8)
    v.annotations["pn7"].add([10, 11], "1", 12)

    assert v._current_idx == 0
    v(simulate_key_press(v.figure, key="p"))  # previous frame - none
    assert v._current_idx == 0
    v(simulate_key_press(v.figure, key="n"))  # next frame
    assert v._current_idx == 5
    v(simulate_key_press(v.figure, key="n"))
    assert v._current_idx == 8
    v(simulate_key_press(v.figure, key="n"))
    assert (
        v._current_idx == 8
    )  # should not move because there are no more annotations for this label

    plt.close(v.figure)


def test_save(video_fname):
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn7"]
    )
    assert not os.path.exists(v.annotations["pn7"].fname)
    v.save()
    assert os.path.exists(v.annotations["pn7"].fname)
    plt.close(v.figure)


def test_video_point_annotator_keep_overlapping_continuous_frames(video_fname):
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn8"]
    )
    v.ann.data = {}
    v.ann.add_label("0")
    v.ann.add_label("1")
    v.ann.add_label("2")

    # add annotations in frame 10-50 for label 0
    for i in range(10, 51):
        v.annotations["pn8"].add([i, i], "0", i)

    # add annotations in frame 20-60 for label 1
    for i in range(20, 61):
        v.annotations["pn8"].add([i, i], "1", i)

    # add annotations in frame 5-55 for label 2
    for i in range(5, 56):
        v.annotations["pn8"].add([i, i], "2", i)

    v(simulate_key_press(v.figure, key="alt+q"))  # keep overlapping continuous frames
    assert len(v.annotations["pn8"].frames) == 31  # all frames should be kept
    plt.close(v.figure)


def test_video_point_annotator_render(video_fname):
    v = datanavigator.VideoPointAnnotator(
        vid_name=video_fname, annotation_names=["pn9"]
    )
    v.annotations["pn9"].add([10, 100], "0", 0)  # add some data
    v.annotations["pn9"].add([50, 200], "0", 60)  # add some data
    out_vid_name = os.path.join(
        os.path.dirname(video_fname), "test_video_point_annotator_render.mp4"
    )
    assert not os.path.exists(out_vid_name)
    v.render(10, 20, out_vid_name)
    assert os.path.exists(out_vid_name)
    # check that the video is rendered correctly
    assert len(datanavigator.Video(out_vid_name)) == 11

    plt.close(v.figure)



from unittest.mock import patch
from datanavigator import _check_ffmpeg, _check_clip_folder


@patch("shutil.which")
@patch("builtins.print")
def test_check_ffmpeg_installed(mock_print, mock_which):
    # Simulate ffmpeg being available in PATH
    mock_which.return_value = "/usr/bin/ffmpeg"
    _check_ffmpeg()
    mock_print.assert_not_called()


@patch("shutil.which")
@patch("builtins.print")
def test_check_ffmpeg_not_installed(mock_print, mock_which):
    # Simulate ffmpeg not being available in PATH
    mock_which.return_value = None
    with patch("sys.platform", "win32"):
        _check_ffmpeg()
        mock_print.assert_any_call("\nFFmpeg is not installed or not in PATH.")
        mock_print.assert_any_call("Download it from: https://ffmpeg.org/download.html")


@patch("os.path.exists")
@patch("builtins.print")
def test_check_clip_folder_exists(mock_print, mock_path_exists):
    # Simulate clip folder already existing
    mock_path_exists.return_value = True
    _check_clip_folder()
    mock_print.assert_not_called()



r"""
Interactive data visualization for signals, videos, and complex data objects.

Browsers

- :py:class:`GenericBrowser`: Generic class to browse data. Meant to be extended.
- :py:class:`SignalBrowser`: Browse an array of pysampled.Data elements, or 2D arrays.
- :py:class:`PlotBrowser`: Scroll through an array of complex data where a plotting function is defined for each element.
- :py:class:`VideoBrowser`: Scroll through the frames of a video.
- :py:class:`VideoPlotBrowser`: Browse through video and 1D signals synced to the video side by side.
- :py:class:`ComponentBrowser`: Browse signals (e.g., from periodic motion) as scatterplots of components (e.g., from UMAP, PCA).


Point tracking

- :py:class:`Video`: Extended VideoReader class with additional functionalities (helper for VideoPointAnnotator).
- :py:class:`VideoAnnotation`: Manage one point annotation layer in a video.
- :py:class:`VideoAnnotations`: Manager for multiple video annotation layers.
- :py:class:`VideoPointAnnotator`: Annotate points in a video.

Optical flow

- :py:func:`lucas_kanade`: Track points in a video using the Lucas-Kanade algorithm.
- :py:func:`lucas_kanade_rstc`: Track points in a video using Lucas-Kanade with reverse sigmoid tracking correction.
- :py:func:`test_lucas_kanade_rstc`: Test function for Lucas-Kanade with reverse sigmoid tracking correction.

Assets

- :py:class:`Button`: Custom button widget with a 'name' state.
- :py:class:`StateButton`: Button widget that stores a number/coordinate state.
- :py:class:`ToggleButton`: Button widget with a toggle state.
- :py:class:`Selector`: Select points in a plot using the lasso selection widget.
- :py:class:`StateVariable`: Manage state variables with multiple states.
- :py:class:`EventData`: Manage the data from one event type in one trial.
- :py:class:`Event`: Manage selection of a sequence of events.

Assetcontainers

- :py:class:`AssetContainer`: Container for managing assets such as buttons, memory slots, etc.
- :py:class:`Buttons`: Manager for buttons in a matplotlib figure or GUI.
- :py:class:`Selectors`: Manager for selector objects for picking points on line2D objects.
- :py:class:`MemorySlots`: Manager for memory slots to store and navigate positions.
- :py:class:`StateVariables`: Manager for state variables.
- :py:class:`Events`: Manager for event objects.
"""
import os
import sys
import shutil

from .__version__ import __version__

from ._config import (
    get_cache_folder,
    get_clip_folder,
    set_cache_folder,
    set_clip_folder,
)
from .assets import (
    AssetContainer,
    Button,
    Buttons,
    MemorySlots,
    Selector,
    Selectors,
    StateButton,
    StateVariable,
    StateVariables,
    ToggleButton,
)
from .events import portion, Event, EventData, Events

from .core import GenericBrowser
from .plots import PlotBrowser
from .signals import SignalBrowser
from .videos import VideoBrowser, VideoPlotBrowser
from .components import ComponentBrowser

from .opticalflow import lucas_kanade, lucas_kanade_rstc
from .pointtracking import VideoAnnotation, VideoAnnotations, VideoPointAnnotator

from .utils import (
    TextView,
    Video,
    get_palette,
    is_path_exists_or_creatable,
    is_video,
    ticks_from_times,
)

from .examples import (
    get_example_video,
    EventPickerDemo,
    ButtonDemo,
    SelectorDemo,
)


def _check_ffmpeg():
    def check_command(command):
        """Check if a command is available in the system's PATH."""
        return shutil.which(command) is not None

    def print_install_instructions():
        """Print installation instructions for ffmpeg and ffprobe."""
        if sys.platform.startswith("win"):
            print("\nFFmpeg is not installed or not in PATH.")
            print("Download it from: https://ffmpeg.org/download.html")
            print("After installation, add FFmpeg's 'bin' folder to the system PATH.")
        else:
            print("\nFFmpeg is not installed or not in PATH.")
            print("On Debian/Ubuntu, install it with: sudo apt install ffmpeg")
            print("On macOS, install it with: brew install ffmpeg")
            print("On Fedora, install it with: sudo dnf install ffmpeg")
            print("On Arch Linux, install it with: sudo pacman -S ffmpeg")

    # Check if ffmpeg and ffprobe are available
    ffmpeg_found = check_command("ffmpeg")

    if not ffmpeg_found:
        print("Cound not find ffmpeg.")
        print_install_instructions()


def _check_clip_folder():
    if not os.path.exists(get_clip_folder()):
        folder = os.getcwd()
        print(f"Using the current working directory-{folder}-for storing video clips.")
        set_clip_folder(folder)
        print("To change, use datanavigator.set_clip_folder(<folder_name>)")


_check_ffmpeg()
_check_clip_folder()
